# SPDX-License-Identifier: BSD-2-Clause
""" Compares two source trees. """

# Copyright (C) 2019, 2025 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import dataclasses
import difflib
import os
import re
from typing import NamedTuple, Union

import specitems

from .directorystate import DirectoryState
from .sphinxbuilder import SphinxBuilder

_COMMENT = re.compile(r"\s*(\*|\/\*|\s*$)")

_WARNING_NO_REVIEW = """.. topic:: WARNING

    This change has no associated review.
"""


@dataclasses.dataclass
class CompareSourcesConfig:
    """ Represents a source comparison configuration. """
    current: DirectoryState
    previous: DirectoryState
    reviews: dict[str, dict[str, str]]
    file_to_review: dict[str, Union[str, list[str]]]
    renamed: dict[str, str]
    label: str


class _Context:
    # pylint: disable=too-many-instance-attributes

    def __init__(self, content: specitems.TextContent,
                 file_to_review: dict[str, Union[str, list[str]]]):
        self.content = content
        self.file_to_review = file_to_review
        self.new_files: set[str] = set()
        self.removed_files: set[str] = set()
        self.other: dict[str, list[list[str]]] = {}
        self.inline: dict[str, list[list[str]]] = {}
        self.comments: dict[str, list[list[str]]] = {}
        self.no_changes: set[str] = set()
        self.review_to_new_files: dict[str, list[str]] = {}
        self.review_to_removed_files: dict[str, list[str]] = {}
        self.review_to_other_files: dict[str, list[str]] = {}
        self.label: str = ""

    def make_label(self, key: str) -> str:
        """ Make a label derived from the key. """
        return specitems.make_label(f"{self.label} {key}")

    def add_review_ref(self, name: str) -> None:
        """
        Add the review reference associated with the name to the content.
        """
        maybe_key = self.file_to_review.get(name, None)
        if maybe_key is None:
            self.content.add(_WARNING_NO_REVIEW)
        elif isinstance(maybe_key, str):
            self.content.add(f"See :ref:`{self.make_label(maybe_key)}`.")
        else:
            self.content.add_list(
                [f":ref:`{self.make_label(key)}`" for key in maybe_key],
                "See the following reviews:")

    def add_diffs(self, diffs: dict[str, list[list[str]]],
                  ref_review: bool) -> None:
        """
        Add the file differences to the content.

        Optionally adds review references.
        """
        for name, chunks in sorted(diffs.items()):
            if ref_review:
                label = self.make_label(name)
            else:
                label = None
            with self.content.section(name, label=label):
                if ref_review:
                    self.add_review_ref(name)
                line_number_start = 1
                for chunk in chunks:
                    self.content.add_code_block(
                        chunk,
                        language="diff",
                        line_number_start=line_number_start)
                    line_number_start += len(chunk)

    def add_files(self, files: set[str], section: str) -> None:
        """ Add the section with files to the content. """
        with self.content.section(section):
            for name in sorted(files):
                with self.content.section(name, label=self.make_label(name)):
                    self.add_review_ref(name)

    def add_related_files(self, key: str, review_to_files: dict[str,
                                                                list[str]],
                          action: str) -> None:
        """
        Add the files related to the review associated with the key to the
        content.
        """
        files = review_to_files.get(key, None)
        if files is None:
            return
        if len(files) == 1:
            self.content.add(f"This review is related to the {action} "
                             f"file :ref:`{self.make_label(files[0])}`.")
        else:
            self.content.add_list(
                (f":ref:`{self.make_label(name)}`" for name in sorted(files)),
                f"This review is related to the following {action} files:")

    def add_to_other_or_inline(self, name: str, chunk: list[str]) -> None:
        """ Add the chunk to the other or inline set. """
        minus = 0
        plus = 0
        inline_minus = 0
        inline_plus = 0
        for line in chunk:
            i = line[0]
            minus += int(i == "-")
            plus += int(i == "+")
            inline_minus += int(i == "-" and "RTEMS_INLINE_ROUTINE" in line)
            inline_plus += int(i == "+" and "static inline" in line)
        if minus == inline_minus and plus == inline_plus:
            self.inline.setdefault(name, []).append(chunk)
        else:
            self.other.setdefault(name, []).append(chunk)


def _get_chunks(prefix_current: str, prefix_previous: str, name_current: str,
                name_previous: str) -> list[list[str]]:
    chunks: list[list[str]] = []
    path = os.path.join(prefix_current, name_current)
    path_previous = os.path.join(prefix_previous, name_previous)
    with open(path,
              encoding="utf-8") as src, open(path_previous,
                                             encoding="utf-8") as src_previous:
        chunk: list[str] = []
        for line in difflib.unified_diff(src_previous.readlines(),
                                         src.readlines(), path_previous, path):
            line = line.rstrip("\r\n")
            if line.startswith("@@"):
                chunks.append(chunk)
                chunk = []
            chunk.append(line)
        if chunk:
            chunks.append(chunk)
    return chunks[1:]


class _Sources(NamedTuple):
    prefix: str
    files: set[str]


def _review_to_files(
        files: set[str],
        file_to_review: dict[str, Union[str,
                                        list[str]]]) -> dict[str, list[str]]:
    review_to_files: dict[str, list[str]] = {}
    for name in files:
        maybe_key = file_to_review.get(name, None)
        if maybe_key is None:
            continue
        if isinstance(maybe_key, str):
            review_to_files.setdefault(maybe_key, []).append(name)
        else:
            for key in maybe_key:
                review_to_files.setdefault(key, []).append(name)
    return review_to_files


def _add_comparison(content: specitems.TextContent,
                    config: CompareSourcesConfig, current: _Sources,
                    previous: _Sources) -> None:
    ctx = _Context(content, config.file_to_review)
    for name_current in current.files:
        name_previous = config.renamed.get(name_current, name_current)
        if name_previous not in previous.files:
            ctx.new_files.add(name_current)
            continue
        previous.files.remove(name_previous)
        chunks = _get_chunks(current.prefix, previous.prefix, name_current,
                             name_previous)
        if not chunks:
            ctx.no_changes.add(name_current)
            continue
        for chunk in chunks:
            for line in chunk:
                if line.startswith(("-", "+")):
                    if _COMMENT.match(line[1:]) is None:
                        ctx.add_to_other_or_inline(name_current, chunk)
                        break
            else:
                ctx.comments.setdefault(name_current, []).append(chunk)
    ctx.removed_files = previous.files
    ctx.review_to_new_files = _review_to_files(ctx.new_files,
                                               ctx.file_to_review)
    ctx.review_to_removed_files = _review_to_files(ctx.removed_files,
                                                   ctx.file_to_review)
    ctx.review_to_other_files = _review_to_files(set(ctx.other.keys()),
                                                 ctx.file_to_review)
    with ctx.content.label_scope(config.label):
        ctx.label = ctx.content.get_label()
        with ctx.content.section("Reviews"):
            for key, review in config.reviews.items():
                with ctx.content.section(review["subject"],
                                         label=ctx.make_label(key)):
                    ctx.content.add(review["text"])
                    ctx.add_related_files(key, ctx.review_to_new_files, "new")
                    ctx.add_related_files(key, ctx.review_to_removed_files,
                                          "removed")
                    ctx.add_related_files(key, ctx.review_to_other_files,
                                          "changed")
        ctx.add_files(ctx.new_files, "New files")
        ctx.add_files(previous.files, "Removed files")
        with ctx.content.section("Files with unspecific changes"):
            ctx.add_diffs(ctx.other, True)
        if ctx.inline:
            with ctx.content.section(
                    "Files with RTEMS_INLINE_ROUTINE changes"):
                ctx.content.add(f"See :ref:`{ctx.make_label('inline')}`.")
                ctx.add_diffs(ctx.inline, False)
        with ctx.content.section("Files with changes inside of comments"):
            ctx.add_diffs(ctx.comments, False)
        with ctx.content.section("Unchanged files"):
            ctx.content.add_list(sorted(ctx.no_changes))


def compare_sources(content: specitems.TextContent,
                    config: CompareSourcesConfig) -> None:
    """ Compare two source trees according to the configuration. """
    sources_current = _Sources(config.current.directory,
                               set(config.current.files(base="")))
    sources_previous = _Sources(config.previous.directory,
                                set(config.previous.files(base="")))
    _add_comparison(content, config, sources_current, sources_previous)


class CompareSourcesProvider(specitems.ItemValueProvider):
    """ Provides a source comparison. """

    # pylint: disable=too-few-public-methods

    def __init__(self, builder: SphinxBuilder) -> None:
        super().__init__(builder.mapper)
        self._builder = builder
        self.mapper.add_get_value(f"{builder.item.type}:/compare-sources",
                                  self._get_compare_sources)

    def _get_compare_sources(self, ctx: specitems.ItemGetValueContext) -> str:
        builder = self._builder
        with builder.section_level_scope(ctx) as key:
            assert key
            current_link, current_sources = builder.input_link_by_key(
                "source-compare-current", "source-compare-key", key)
            assert isinstance(current_sources, DirectoryState)
            _, previous_sources = builder.input_link_by_key(
                "source-compare-previous", "source-compare-key", key)
            assert isinstance(previous_sources, DirectoryState)
            current_data = ctx.mapper.substitute_data(current_link.data,
                                                      builder.item)
            config = CompareSourcesConfig(
                current=current_sources,
                previous=previous_sources,
                reviews=current_data["reviews"],
                file_to_review=current_data["file-to-review"],
                renamed=current_data["renamed"],
                label=current_data["label"])
            assert isinstance(self.mapper, specitems.TextMapper)
            content = self.mapper.create_content(
                section_level=builder.section_level)
            compare_sources(content, config)
            return content.join()



# SPDX-License-Identifier: BSD-2-Clause
""" Builds a package manual. """

# Copyright (C) 2020, 2025 embedded brains GmbH & Co. KG
# Copyright (C) 2021 EDISOFT
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import json
import logging
import os
from typing import Iterator

from specitems import (COL_SPAN, Copyrights, Item, ItemGetValueContext,
                       make_label, Link, SphinxContent, SphinxMapper)
from specware import (BSD_2_CLAUSE_LICENSE, MIT_LICENSE, gather_api_items,
                      run_command)

from .archiver import Archiver
from .directorystate import DirectoryState, RepositoryState
from .docbuilder import DocumentBuilder
from .membench import generate, generate_variants_table, MembenchVariant
from .pkgitems import PackageBuildDirector, PackageComponent
from .packagechanges import PackageChanges
from .testaggregator import TestAggregator
from .testrunner import TestLog

_EnvToStats = dict[str, tuple[float, float, float]]
_ItemToEnvStats = dict[str, _EnvToStats]

_LICENSE_LISTING = {"BSD-2-Clause": BSD_2_CLAUSE_LICENSE, "MIT": MIT_LICENSE}


def _run_pkg_config(content: SphinxContent, cmd: list[str]) -> None:
    cmd = ["pkg-config"] + cmd
    content.add(f"$ {' '.join(cmd)}")
    stdout: list[str] = []
    status = run_command(cmd, stdout=stdout)
    assert status == 0
    content.append(stdout)


def _gather_runtime_performance_items(items: set[Item], item: Item) -> None:
    if item.type == "requirement/non-functional/performance-runtime":
        items.add(item)
    for child in item.children("requirement-refinement"):
        _gather_runtime_performance_items(items, child)


def _environment_order(name: str) -> int:
    if name == "HotCache":
        return 0
    if name == "FullCache":
        return 1
    if name == "DirtyCache":
        return 2
    return int(name[5:]) + 2


def _add_licenses(content: SphinxContent, deployment_directory: str,
                  member: DirectoryState,
                  license_listing: dict[str, Copyrights]) -> None:
    copyrights_by_license = member["copyrights-by-license"]
    files = copyrights_by_license.get("files", None)
    if files is not None:
        directory = os.path.relpath(member.directory, deployment_directory)
        with content.section(f"Directory - {directory}"):
            content.add(copyrights_by_license.get("description", None))
            for name in files:
                with content.section(f"File - {name}"):
                    content.add(f"""The license file
{content.path(os.path.join(directory, name))}
is applicable to this directory or parts of the directory:""")
                    content.add_blank_line()
                    file_path = os.path.join(member.directory, name)
                    with open(file_path, "r", encoding="utf-8") as src:
                        content.add_code_block(src.readlines(),
                                               line_number_start=-1)
    for the_license in license_listing:
        license_listing[the_license].register(
            copyrights_by_license.get(the_license, []))


def _get_performance_environments(
        measurements_by_variant: dict[int, _ItemToEnvStats]) -> list[str]:
    # We cannot use the performance runtime measurement environments specified
    # by items, since we have a target defined number of the Load/N
    # environments.
    #
    # Iterate over all gathered performance statistics since some providers may
    # have no results.  Try the best to produce a report even with insufficient
    # data.
    all_envs: set[str] = set()
    for item_to_env_stats in measurements_by_variant.values():
        for env_to_stats in item_to_env_stats.values():
            all_envs.update(env_to_stats.keys())
    return sorted(all_envs, key=_environment_order)


def _gather_targets(component: PackageComponent, targets: set[str]) -> None:
    with component.item.cache.selection(component.selection):
        targets.update(item.uid
                       for item in component.item.parents("design-target"))
    for child in component.item.children("input"):
        if child.type.startswith("pkg/component"):
            _gather_targets(component.director[child.uid], targets)


class PackageManualBuilder(DocumentBuilder):
    """ Builds a package user manual. """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        self._membench: dict[str, dict] = {}
        my_type = self.item.type
        self.mapper.add_get_value(f"{my_type}:/benchmark-variants-list",
                                  self._get_benchmark_variants_list)
        self.mapper.add_get_value(f"{my_type}:/memory-benchmarks",
                                  self._get_membench)
        self.mapper.add_get_value(f"{my_type}:/memory-benchmark-section",
                                  self._get_membench_section)
        self.mapper.add_get_value(f"{my_type}:/memory-benchmark-reference",
                                  self._get_membench_ref)
        self.mapper.add_get_value(
            f"{my_type}:/memory-benchmark-based-on-reference",
            self._get_membench_based_on_ref)
        self.mapper.add_get_value(
            f"{my_type}:/memory-benchmark-compare-section",
            self._get_membench_compare_section)
        self.mapper.add_get_value(
            f"{my_type}:/memory-benchmark-variants-table",
            self._get_membench_variants_table)
        self.mapper.add_get_value(f"{my_type}:/performance-variants-table",
                                  self._get_performance_variants_table)
        self.mapper.add_get_value(f"{my_type}:/pkg-config",
                                  self._get_pkg_config)
        self.mapper.add_get_value(f"{my_type}:/pre-qualified-interfaces",
                                  self._get_pre_qualified_interfaces)
        self.mapper.add_get_value(f"{my_type}:/repositories", self._get_repos)
        self.mapper.add_get_value(f"{my_type}:/change-list",
                                  self._get_change_list)
        self.mapper.add_get_value(f"{my_type}:/open-issues",
                                  self._get_open_issues)
        self.mapper.add_get_value(f"{my_type}:/license-info",
                                  self._get_license_info)
        self.mapper.add_get_value(f"{my_type}:/targets", self._get_targets)
        for name in [my_type, "pkg/sphinx-section"]:
            self.mapper.add_get_value(f"{name}:/object-size",
                                      self._get_object_size)

    def run(self) -> None:
        for membench in self.inputs("membench-results"):
            assert isinstance(membench, DirectoryState)
            self._membench.update(membench.json_load())
        super().run()

    def _get_benchmark_variants_list(self, ctx: ItemGetValueContext) -> str:
        with self.section_level_scope(ctx):
            content = SphinxContent(section_level=self.section_level)
            for link, test_log in self._yield_benchmark_variants():
                content.add_definition_item(
                    test_log.substitute(link["variant-name"]),
                    test_log.substitute(link["description"]))
            return content.join()

    def _get_membench_build_label(self) -> str:
        return self.substitute("${.:/component/membench-build-label}")

    def _get_membench(self, ctx: ItemGetValueContext) -> str:
        with self.section_level_scope(ctx):
            content = SphinxContent(section_level=self.section_level)
            content.push_label(
                make_label(self.substitute("${.:/component/ident}")))
            label = self._get_membench_build_label()
            sections_by_uid = self._membench[label]["membench"]
            root = self.item.cache["/rtems/req/mem-basic"]
            table_pivots = ["/rtems/req/mem-smp-1"]
            generate(content, sections_by_uid, root, table_pivots,
                     SphinxMapper(root))
            return content.join()

    def _get_membench_section(self, ctx: ItemGetValueContext) -> str:
        assert ctx.args
        uid, section = ctx.args.split(":")
        item = self.item.cache[uid]
        label = self._get_membench_build_label()
        sections = self._membench[label]["membench"][item.uid]
        return str(sections[section])

    def _membench_get_ref(self, ctx: ItemGetValueContext, section: str) -> str:
        assert ctx.args
        item = self.item.cache[ctx.args]
        scope = make_label(self.substitute("${.:/component/ident}"))
        label = f"{scope}{section}{item.ident}"
        return f":ref:`{item.spec_2} <{label}>`"

    def _get_membench_ref(self, ctx: ItemGetValueContext) -> str:
        return self._membench_get_ref(ctx, "BenchmarkSpec")

    def _get_membench_based_on_ref(self, ctx: ItemGetValueContext) -> str:
        return self._membench_get_ref(ctx, "BenchmarksBasedOnSpec")

    def _get_membench_compare_section(self, ctx: ItemGetValueContext) -> str:
        assert ctx.args
        uid, other_uid, section = ctx.args.split(":")
        item = self.item.cache[uid]
        label = self._get_membench_build_label()
        sections = self._membench[label]["membench"][item.uid]
        other_item = self.item.cache[other_uid]
        other_sections = self._membench[label]["membench"][other_item.uid]
        return f"{other_sections[section] - sections[section]:+}"

    def _yield_benchmark_variants(self) -> Iterator[tuple[Link, TestLog]]:
        for link, test_log in sorted(self.input_links("benchmark-variant"),
                                     key=lambda x: x[0]["priority"]):
            logging.info("%s: use benchmark variant of: %s", self.uid,
                         test_log.uid)
            assert isinstance(test_log, TestLog)
            yield link, test_log

    def _get_membench_variants_table(self, ctx: ItemGetValueContext) -> str:
        with self.section_level_scope(ctx):
            content = SphinxContent(section_level=self.section_level)
            root = self.item.cache["/rtems/req/mem-basic"]
            variants: list[MembenchVariant] = []
            for link, test_log in self._yield_benchmark_variants():
                build_label = test_log.input(
                    "build-configuration")["build-label"]
                if build_label not in self._membench:
                    logging.warning(
                        "%s: no memory benchmark for build label: %s",
                        self.uid, build_label)
                    continue
                variants.append(
                    MembenchVariant(test_log.substitute(link["variant-name"]),
                                    build_label))
            generate_variants_table(content, self._membench, root, variants)
            return content.join()

    def _get_measurements_by_variant(self) -> dict[int, _ItemToEnvStats]:
        measurements_by_variant: dict[int, _ItemToEnvStats] = {}
        for index, (_,
                    test_log) in enumerate(self._yield_benchmark_variants()):
            with open(test_log.file, "r", encoding="utf-8") as src:
                data = json.load(src)
            measurements: _ItemToEnvStats = {}
            for report in data["reports"]:
                for test_case in report.get("test-suite",
                                            {}).get("test-cases", []):
                    for measurement in test_case["runtime-measurements"]:
                        stats = (measurement["min"], measurement["q2"],
                                 measurement["max"])
                        measurements.setdefault(
                            measurement["name"],
                            {})[measurement["variant"]] = stats
            measurements_by_variant[index] = measurements
        return measurements_by_variant

    def _make_performance_variants_rows(
        self, item: Item, envs: list[str], env_links: dict[str, str],
        measurements_by_variant: dict[int, _ItemToEnvStats]
    ) -> list[tuple[str | int, ...]]:
        # pylint: disable=too-many-locals
        rows: list[tuple[str | int, ...]] = []
        info_spec: str | int = self.mapper.get_link(item)
        for env in envs:
            env_link = env_links[env.split("/")[0]]
            info_env: str | int = f"`{env} <{env_link}>`__"
            for index, (link, test_log) in enumerate(
                    self._yield_benchmark_variants()):
                stats = measurements_by_variant[index].get(item.ident,
                                                           {}).get(env, None)
                info = (info_spec, info_env,
                        test_log.substitute(link["variant-name"]))
                if index == 0:
                    if not stats:
                        rows.append(info + ("?", "?", "?"))
                        continue
                    base = stats
                    rows.append(info + tuple(f"{value * 1e6:.3f}"
                                             for value in stats))
                elif stats:
                    rows.append(info + tuple(
                        f"{(value - base[j]) / base[j] * 100.0:+.3g} %"
                        for j, value in enumerate(stats)))
                else:
                    rows.append(info + ("?", "?", "?"))
                info_spec = COL_SPAN
                info_env = COL_SPAN
        return rows

    def _get_performance_variants_table(self, ctx: ItemGetValueContext) -> str:
        with self.section_level_scope(ctx):
            measurements_by_variant = self._get_measurements_by_variant()
            if not measurements_by_variant:
                return "There is no performance variants table available."
            items: set[Item] = set()
            _gather_runtime_performance_items(items,
                                              self.item.cache["/req/root"])
            _, test_log = next(self._yield_benchmark_variants())
            envs = _get_performance_environments(measurements_by_variant)
            req_path = test_log.substitute(
                "${.:/component/deployment-directory}/doc/ts/srs/"
                "html/requirements.html#spec-req-perf-runtime-environment-")
            env_links = {
                "HotCache": f"{req_path}hot-cache",
                "FullCache": f"{req_path}full-cache",
                "DirtyCache": f"{req_path}dirty-cache",
                "Load": f"{req_path}load"
            }
            rows: list[tuple[str | int, ...]] = [
                ("Specification", "Environment", "Variant", "Min [μs]",
                 "Median [μs]", "Max [μs]")
            ]
            for item in sorted(items):
                rows.extend(
                    self._make_performance_variants_rows(
                        item, envs, env_links, measurements_by_variant))
            content = SphinxContent(section_level=self.section_level)
            content.add_grid_table(rows, [31, 14, 19, 12, 12, 12],
                                   font_size=-3)
            return content.join()

    def _get_pkg_config(self, _ctx: ItemGetValueContext) -> str:
        pkg = self.substitute(
            "${.:/component/deployment-directory}/lib/pkgconfig/"
            "${.:/component/arch}-rtems${.:/component/rtems-version}-"
            "${.:/component/bsp}${.:/component/config:dash}"
            "${.:/component/bsp-qual-only:dash}.pc")
        content = SphinxContent()
        with content.directive("code-block", value="none"):
            _run_pkg_config(content, ["--variable=ABI_FLAGS", pkg])
            _run_pkg_config(content, ["--cflags", pkg])
            _run_pkg_config(content, ["--variable=LDFLAGS", pkg])
        return content.join()

    def _get_pre_qualified_interfaces(self, ctx: ItemGetValueContext) -> str:
        with self.section_level_scope(ctx):
            content = SphinxContent(section_level=self.section_level)
            content.push_label(
                make_label(self.substitute("${.:/component/ident}")))
            items: dict[str, list[Item]] = {}
            gather_api_items(self.item.cache, items)
            for group, group_items in sorted(items.items()):
                with content.section(group):
                    content.add_list(
                        self.mapper.get_link(item) for item in group_items)
            return content.join()

    def _get_repos(self, ctx: ItemGetValueContext) -> str:
        with self.section_level_scope(ctx):
            content = SphinxContent(section_level=self.section_level)
            prefix = self.substitute("${.:/component/deployment-directory}")
            for item in self.component.item.children("repository"):
                repo = self.director[item.uid]
                assert isinstance(repo, RepositoryState)
                assert repo.lazy_verify()
                origin_branch = repo["origin-branch"]
                origin_commit = repo["origin-commit"]
                if origin_branch and origin_commit:
                    dest = os.path.relpath(repo["directory"], prefix)
                    with content.section(f"Git Repository: {dest}"):
                        content.add(repo["description"])
                        content.add(f"""The ``{repo["branch"]}`` branch with
commit ``{repo["commit"]}``
was used to build the package.  This branch is checked out after unpacking the
archive.  It is based on
commit `{origin_commit} <{repo['origin-commit-url']}>`_
of the ``{origin_branch}`` branch of the ``origin`` remote repository.""")
            return content.join()

    def _get_object_size(self, ctx: ItemGetValueContext) -> str:
        assert ctx.args
        label = self._get_membench_build_label()
        return str(self._membench[label]["object-sizes"][ctx.args])

    def _get_change_list(self, ctx: ItemGetValueContext) -> str:
        with self.section_level_scope(ctx):
            changes = self.input("package-changes")
            assert isinstance(changes, PackageChanges)
            return changes.get_change_list(self.mapper, self.section_level)

    def _get_open_issues(self, ctx: ItemGetValueContext) -> str:
        with self.section_level_scope(ctx):
            changes = self.input("package-changes")
            assert isinstance(changes, PackageChanges)
            return changes.get_open_issues(self.mapper, self.section_level)

    def _get_license_info(self, ctx: ItemGetValueContext) -> str:
        archiver = self.input("archive")
        assert isinstance(archiver, Archiver)
        with self.section_level_scope(ctx):
            content = SphinxContent(section_level=self.section_level)
            license_listing = {
                "BSD-2-Clause": Copyrights(),
                "MIT": Copyrights()
            }
            deployment_directory = self.substitute(
                "${.:/component/deployment-directory}")
            content.add(f"""All directories and file paths in this section are
relative to {content.path(deployment_directory)}.""")
            for member in archiver.inputs("member"):
                assert isinstance(member, DirectoryState)
                _add_licenses(content, deployment_directory, member,
                              license_listing)
            for the_license, copyrights in license_listing.items():
                if copyrights:
                    with content.section(f"{the_license} copyrights"):
                        content.add(copyrights.get_statements("| ©"))
                        content.add_code_block(
                            _LICENSE_LISTING[the_license].split("\n"),
                            line_number_start=-1)
            return content.join()

    def _get_targets(self, ctx: ItemGetValueContext) -> str:
        with self.section_level_scope(ctx):
            content = SphinxContent(section_level=self.section_level)
            targets: set[str] = set()
            _gather_targets(self.component, targets)
            for uid in sorted(targets):
                target = self.item.cache[uid]
                with content.section(target["name"], label=self.label(target)):
                    content.add(self.substitute(target["brief"]))
                    content.add(self.substitute(target["description"]))
            return content.join()


class PackageSummary(DirectoryState):
    """ Builds a package summary. """

    def run(self) -> None:
        package_manual = self.input("source")
        content = SphinxContent(section_level=0)
        with content.section(f"Package Summary - {self.component['ident']}"):
            with content.section("Unexpected Test Failures"):
                for item in self.item.cache.items_by_type.get(
                        "pkg/directory-state/sphinx/test-report", []):
                    ident = self.substitute(
                        f"${{{item.uid}:/component/ident}}")
                    with content.section(f"Component - {ident}"):
                        for target, failures in item[
                                "unexpected-test-failures"].items():
                            content.add_list_item(target)
                            content.add_blank_line()
                            with content.indent("  "):
                                content.add_list(failures)
            with content.section("Code/Branch Coverage"):
                self._add_coverage_achievement(content)
            with content.section("Repositories"):
                prefix = package_manual.substitute(
                    "${.:/component/deployment-directory}")
                for item in package_manual.component.item.children(
                        "repository"):
                    repo = self.director[item.uid]
                    repo_dir = repo["directory"]
                    with content.section(os.path.relpath(repo_dir, prefix)):
                        stdout: list[str] = []
                        status = run_command(["git", "log", "-1"], repo_dir,
                                             stdout)
                        assert status == 0
                        content.add_code_block(stdout)
        content.write(self.file)
        self.description.add(f"""Produce the package summary file
{content.path(self.file)}.""")

    def _add_coverage_achievement(self, content: SphinxContent) -> None:
        for test_aggregator in self.inputs("test-aggregation"):
            assert isinstance(test_aggregator, TestAggregator)
            ident = test_aggregator.substitute("${.:/component/ident}")
            with content.section(f"Component - {ident}"):
                test_aggregator.add_coverage_achievement(content)



# SPDX-License-Identifier: BSD-2-Clause
""" Provides an item value provider for standard tailoring. """

# Copyright (C) 2025 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
from typing import Any

from specitems import (Item, ItemGetValueContext, ItemValueProvider,
                       TextContent, TextMapper)

from .sphinxbuilder import SphinxBuilder
from .pkgitems import BuildItem, BuildItemFactory, export_data


def _get_ecss_standard_and_clause(ctx: ItemGetValueContext) -> str:
    item = ctx.item
    standard = item.parent("requirement-refinement")["name"]
    clause = f"{item['section']}{item['bullet']}"
    return f":ref:`{standard} {clause} <{item.ident}>`"


def _get_ecss_clause(ctx: ItemGetValueContext) -> str:
    item = ctx.item
    clause = f"{item['section']}{item['bullet']}"
    return f":ref:`{clause} <{item.ident}>`"


def _add_compliance_matrix(content: TextContent, name: str,
                           clauses: list[Item]) -> None:
    table_name = f"Compliance matrix of {name}"
    options = [":widths: 13, 12, 13, 12, 13, 12, 13, 12", ":header-rows: 1"]
    with content.directive("list-table", table_name, options):
        content.add("""* - Clause
  - Status
  - Clause
  - Status
  - Clause
  - Status
  - Clause
  - Status""")
        for index, clause in enumerate(clauses):
            star = "*" if index % 4 == 0 else " "
            link = clause.child_link("statement-of-compliance")
            the_clause = f"{clause['section']}{clause['bullet']}"
            content.append(f"""{star} - {the_clause}
  - .. _CM{clause.ident}:

    :ref:`{link['supplier-status']} <{clause.ident}>`""")
        odd = len(clauses) % 4
        missing = 4 - odd if odd else 0
        content.append(missing * ["  - \u200b", "  - \u200b"])


def _add_topic(content: TextContent, builder: SphinxBuilder, clause: Item,
               topic: str, text: str | None) -> None:
    if text:
        with content.directive("topic", topic):
            content.add(builder.substitute(text, clause))


def _add_clause(content: TextContent, builder: SphinxBuilder, name: str,
                clause: Item) -> None:
    link = clause.child_link("statement-of-compliance")
    the_clause = f"{clause['section']}{clause['bullet']}"
    header = f"{clause['name']} ({the_clause})"
    with content.section(header, label=clause.ident):
        ident = f"{name[:12]}_{os.path.basename(clause.uid)}"
        _add_topic(content, builder, clause,
                   f"{name} - Clause {the_clause} - {ident}", clause["text"])
        _add_topic(content, builder, clause, "Aim", clause["aim"])
        if len(clause["notes"]) == 1:
            _add_topic(content, builder, clause, "Note", clause["notes"][0])
        else:
            for index, note in enumerate(clause["notes"]):
                _add_topic(content, builder, clause, f"Note {index + 1}", note)
        _add_topic(content, builder, clause, "Expected Output",
                   clause["expected-output"])
        content.add_definition_item(
            f"Tailoring Status: {link['supplier-status']}",
            builder.substitute(link.item["text"], link.item))
        content.add("For an overview of all clauses, see the "
                    f":ref:`tailoring table <CM{clause.ident}>`.")


def _ecss_order(item: Item) -> list:
    return [int(i) for i in item["section"].split(".")] + [item["bullet"]]


class StandardTailoringProvider(ItemValueProvider):
    """ Provides a statement of compliance and a standard tailoring. """

    # pylint: disable=too-few-public-methods

    def __init__(self, builder: SphinxBuilder) -> None:
        super().__init__(builder.mapper)
        self._builder = builder
        mapper = builder.mapper
        for name in [builder.item.type, "pkg/sphinx-section"]:
            mapper.add_get_value(f"{name}:/standard-tailoring",
                                 self._get_standard_tailoring)
        mapper.add_get_value("requirement/non-functional/ecss:/clause",
                             _get_ecss_clause)
        mapper.add_get_value(
            "requirement/non-functional/ecss:/standard-and-clause",
            _get_ecss_standard_and_clause)

    def _get_standard_tailoring(self, ctx: ItemGetValueContext) -> str:
        builder = self._builder
        with builder.section_level_scope(ctx):
            assert isinstance(self.mapper, TextMapper)
            content = self.mapper.create_content(
                section_level=builder.section_level)
            for std in builder.item.parents("standard-tailoring"):
                name = builder.substitute(std["name"], std)
                with content.section(f"Tailoring of {name}", label=std.ident):
                    clauses = list(
                        sorted(std.children("requirement-refinement"),
                               key=_ecss_order))
                    _add_compliance_matrix(content, name, clauses)
                    for clause in clauses:
                        _add_clause(content, builder, name, clause)
            return content.join()


def _export_data(item: Item, present: bool, built_later: bool) -> Any:
    data = export_data(item, present, built_later)
    for key in ("aim", "expected-output", "notes"):
        data[key] = None
    data["name"] = "Text removed due to license issues"
    data["text"] = "Text removed due to license issues."
    return data


class ECSSClause(BuildItem):
    """ Controls the ECSS data export. """

    @classmethod
    def prepare_factory(cls, factory: BuildItemFactory,
                        type_name: str) -> None:
        BuildItem.prepare_factory(factory, type_name)
        factory.add_export_data(type_name, _export_data)



# SPDX-License-Identifier: BSD-2-Clause
""" Maintains a directory state. """

# Copyright (C) 2020, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import base64
import fnmatch
import hashlib
import json
import logging
import os
from pathlib import Path
import re
import shutil
import stat
import tarfile
from typing import Any, Callable, Iterable, Iterator, Optional

from specitems import (Item, ItemGetValueContext, ItemValueProvider,
                       base64_to_hex_text, hash_file)

from .pkgitems import (BuildItem, BuildItemFactory, BuildItemMapper,
                       PackageBuildDirector, export_data)

_Path = Path | str


def _get_file_path(ctx: ItemGetValueContext) -> str:
    index = max(ctx.index, 0)
    return ctx.substitute_and_transform(
        f"{ctx.item['directory']}/{ctx.item['files'][index]['file']}")


def _get_file_path_without_extension(ctx: ItemGetValueContext) -> str:
    return os.path.splitext(_get_file_path(ctx))[0]


def _get_sha512(ctx: ItemGetValueContext) -> str:
    index = max(ctx.index, 0)
    digest = ctx.item["files"][index]["hash"]
    assert digest is not None
    return base64_to_hex_text(digest)


def _do_glob(ctx: ItemGetValueContext) -> list[str]:
    files: list[str] = []
    directory = ctx.mapper.substitute(ctx.item["directory"])
    include_patterns = ctx.unpack_args_list()
    for file_info in ctx.item["files"]:
        path = os.path.join("/", file_info["file"])
        for pattern in include_patterns:
            if fnmatch.fnmatch(path, pattern):
                files.append(os.path.join(directory, file_info["file"]))
                break
    return sorted(files)


def _get_glob(ctx: ItemGetValueContext) -> list[str]:
    files = _do_glob(ctx)
    logging.info("%s: glob files: %s", ctx.item.uid, files)
    return files


def _get_glob_executables(ctx: ItemGetValueContext) -> list[str]:
    files: list[str] = []
    executable = stat.S_IXUSR
    for path in _do_glob(ctx):
        if os.stat(path).st_mode & executable:
            files.append(path)
    logging.info("%s: glob executable files: %s", ctx.item.uid, files)
    return files


def _file_nop(_source: _Path, _target: _Path) -> None:
    pass


def _export_data(item: Item, present: bool, built_later: bool) -> Any:
    data = export_data(item, present, built_later)
    if not present or built_later:
        if "patterns" in item:
            data["files"] = []
        else:
            for file_info in data["files"]:
                file_info["hash"] = None
    if built_later:
        data["hash"] = None
    return data


class DirectoryStateBase(BuildItem):
    """ Maintains a directory state. """

    # pylint: disable=too-many-public-methods
    @classmethod
    def prepare_factory(cls, factory: BuildItemFactory,
                        type_name: str) -> None:
        BuildItem.prepare_factory(factory, type_name)
        factory.add_get_value(f"{type_name}:/file", _get_file_path)
        factory.add_get_value(f"{type_name}:/file-without-extension",
                              _get_file_path_without_extension)
        factory.add_get_value(f"{type_name}:/sha512", _get_sha512)
        factory.add_get_value(f"{type_name}:/glob", _get_glob)
        factory.add_get_value(f"{type_name}:/glob-executables",
                              _get_glob_executables)
        factory.add_export_data(type_name, _export_data)

    def __init__(self,
                 director: PackageBuildDirector,
                 item: Item,
                 mapper: Optional[BuildItemMapper] = None) -> None:
        super().__init__(director, item, mapper)
        DirectoryStateValueProvider(self)
        self._discarded_files: set[str] = set()
        self._files: dict[str, Optional[str]] = dict(
            (file_info["file"], file_info["hash"])
            for file_info in item.get("files", []))
        if self.item.get("directory-state-load-before-use", False):
            self.load()

    def __iter__(self):
        yield from self.files()

    @property
    def directory(self) -> str:
        """ Is the base directory of the directory state. """
        return self["directory"]

    @property
    def file(self) -> str:
        """ Is the path of the first file of the directory state. """
        return next(self.files())

    def digest(self) -> str:
        the_digest = self.item["hash"]
        if the_digest is None:
            return super().digest()
        return the_digest

    def _get_hash(self, _base: str, relative_file_path: str) -> str:
        digest = self._files[relative_file_path]
        assert digest is not None
        return digest

    def _hash_file(self, base: str, relative_file_path: str) -> str:
        file_path = os.path.join(base, relative_file_path)
        digest = hash_file(file_path)
        self._files[relative_file_path] = digest
        return digest

    def _add_hashes(self, base: str, hash_file_handler: Callable[[str, str],
                                                                 str]) -> str:
        overall_hash = hashlib.sha512()
        overall_hash.update(base.encode("utf-8"))
        for relative_file_path in sorted(self._files):
            digest = hash_file_handler(base, relative_file_path)
            overall_hash.update(relative_file_path.encode("utf-8"))
            overall_hash.update(digest.encode("ascii"))
        for associate in self.inputs("associate"):
            overall_hash.update(associate.digest().encode("ascii"))
        self._update_item_files()
        digest = base64.urlsafe_b64encode(
            overall_hash.digest()).decode("ascii")
        self.item["hash"] = digest
        return digest

    def _directory_state_exclude(self, base: str, files: set[str]) -> None:
        for exclude_item in self.item.parents("directory-state-exclude"):
            exclude_state = self.director[exclude_item.uid]
            assert isinstance(exclude_state, DirectoryStateBase)
            exclude_files = files.intersection(
                os.path.relpath(path, base) for path in exclude_state)
            logging.info(
                "%s: exclude files of directory state %s: %s", self.uid,
                exclude_item.uid,
                [os.path.join(base, path) for path in sorted(exclude_files)])
            files.difference_update(exclude_files)

    def _load_from_patterns(self, base: str, patterns: Any) -> None:
        """
        Load the files of the directory state according to the pattern
        definition.

        String patterns, whether given directly or in a list, are treated as
        include patterns.  For dictionary patterns, ``include`` must be
        specified explicitly, while ``exclude`` is optional.  The following
        pattern variants are supported:

        .. code-block:: yaml

            patterns: a.b

        .. code-block:: yaml

            patterns:
              - a.b
              - c.d

        .. code-block:: yaml

            patterns:
              exclude:
                - a.b
                - c.d
              include: e.f

        .. code-block:: yaml

            patterns:
              exclude:
                - a.b
                - c.d
              include:
                - e.f
                - g.h

        .. code-block:: yaml

            patterns:
              - exclude: []
                include: a.b
              - include:
                  - c.d
                  - e.f
        """
        logging.info("%s: load pattern defined directory state: %s", self.uid,
                     base)
        files: set[str] = set()
        base_path = Path(base)
        if isinstance(patterns, (dict, str)):
            patterns = (patterns, )
        for include_exclude in patterns:
            more: set[str] = set()
            if isinstance(include_exclude, str):
                includes: Iterable[str] | str = include_exclude
                excludes: Iterable[str] = tuple()
            else:
                includes = include_exclude["include"]
                excludes = include_exclude.get("exclude", tuple())
            if isinstance(includes, str):
                includes = (includes, )
            for include in includes:
                logging.info("%s: add files matching '%s' in: %s", self.uid,
                             include, base)
                more.update(
                    os.path.relpath(path, base)
                    for path in base_path.glob(include)
                    if path.is_symlink() or not path.is_dir())
            for exclude in excludes:
                exclude_files = set(
                    path for path in more
                    if fnmatch.fnmatch(os.path.join("/", path), exclude))
                logging.info("%s: exclude files for pattern '%s': %s",
                             self.uid, exclude, [
                                 os.path.join(base, path)
                                 for path in sorted(exclude_files)
                             ])
                more.difference_update(exclude_files)
            files.update(more)
        self._directory_state_exclude(base, files)
        self._files = dict.fromkeys(files, None)

    def load(self) -> str:
        """ Load the directory state and returns the overall hash. """
        base = self.directory
        patterns = self.get("patterns", None)
        if patterns is not None:
            self._load_from_patterns(base, patterns)
        else:
            logging.info("%s: load explicit directory state: %s", self.uid,
                         base)
        return self._add_hashes(base, self._hash_file)

    def lazy_load(self) -> str:
        """
        Load the directory state if the overall hash is not present and
        returns the overall hash.
        """
        digest = self.item["hash"]
        if digest is not None:
            return digest
        return self.load()

    def files(self, base: Optional[str] = None) -> Iterator[str]:
        """ Yield the file paths of the directory state. """
        if base is None:
            base = self.directory
        for file_path in sorted(self._files):
            yield os.path.join(base, file_path)

    def files_and_hashes(
            self,
            base: Optional[str] = None) -> Iterator[tuple[str, Optional[str]]]:
        """ Yield the file paths and hashes of the directory state. """
        if base is None:
            base = self.directory
        for file_path, file_hash in sorted(self._files.items()):
            yield os.path.join(base, file_path), file_hash

    def compact(self) -> None:
        """
        Remove the common prefix from the files and adds it to the base
        directory.
        """
        prefix = os.path.commonprefix(list(self._files.keys())).rstrip("/")
        if prefix and not os.path.isabs(prefix):
            self.item["directory"] = os.path.join(self.item["directory"],
                                                  prefix)
            self.item["hash"] = None
            self._files = dict(
                (os.path.relpath(path, prefix), None) for path in self._files)
            self._update_item_files()

    def _update_item_files(self):
        self.item["files"] = list({
            "file": path,
            "hash": digest
        } for path, digest in sorted(self._files.items()))

    def clear(self) -> None:
        """ Clear the file set of the directory state. """
        logging.info("%s: clear directory state", self.uid)
        self.item["hash"] = None
        self._files.clear()
        self._update_item_files()

    def invalidate(self) -> None:
        """ Invalidate the directory state. """
        logging.info("%s: invalidate directory state", self.uid)
        self.item["hash"] = None
        if "patterns" in self:
            self._files.clear()
        else:
            self._files = dict.fromkeys(self._files.keys(), None)
        self._update_item_files()

    def remove_files(self) -> list[str]:
        """ Remove the files of the directory state. """
        removed_files: list[str] = []
        if not self._files:
            return removed_files
        base = self.directory
        for file in sorted(self._files):
            path = os.path.join(base, file)
            logging.info("%s: remove: %s", self.uid, path)
            try:
                os.remove(path)
            except FileNotFoundError:
                if "patterns" in self:
                    logging.warning("%s: file not found: %s", self.uid, path)
                else:
                    logging.debug("%s: file not found: %s", self.uid, path)
            else:
                removed_files.append(file)
        return removed_files

    def add_files(self, files: Iterable[_Path]) -> None:
        """ Add the files to the file set of the directory state. """
        self.item["hash"] = None
        more = set(os.path.normpath(name) for name in files)
        self._directory_state_exclude(self.directory, more)
        self._files.update(dict.fromkeys(more, None))
        self._update_item_files()

    def set_files(self, files: Iterable[_Path]) -> None:
        """ Set the file set of the directory state to the files. """
        self.clear()
        self.add_files(files)

    def _copy_file(self, source: _Path, target: _Path) -> None:
        logging.info("%s: copy '%s' to '%s'", self.uid, source, target)
        os.makedirs(os.path.dirname(target), exist_ok=True)
        shutil.copy2(source, target, follow_symlinks=False)

    def _move_file(self, source: _Path, target: _Path) -> None:
        logging.info("%s: move '%s' to '%s'", self.uid, source, target)
        os.makedirs(os.path.dirname(target), exist_ok=True)
        os.replace(source, target)

    def copy_file(self, source: _Path, target: _Path) -> None:
        """
        Copy the file from the source path to the target path.

        Adds the target file to the file set of the directory state.  The
        target path is relative to the base directory of the directory state.
        """
        self._copy_file(source, os.path.join(self.directory, target))
        self.add_files([target])

    def copy_files(self,
                   root_dir: _Path,
                   files: Iterable[_Path],
                   prefix: _Path = ".") -> None:
        """
        Copy the files relative to the root directory to the base directory
        of the directory state using the prefix.

        The base directory of the directory state and the prefix is prepended
        to the file path for each file before it is added to the directory
        state.  Adds the target files to the file set of the directory state.
        """
        file_list: list[str] = []
        base = self.directory
        for name in files:
            file_source = os.path.join(root_dir, name)
            file_list_path = os.path.join(prefix, name)
            file_list.append(file_list_path)
            file_target = os.path.join(base, file_list_path)
            self._copy_file(file_source, file_target)
        self.add_files(file_list)

    def _add_tree(self,
                  root_dir: _Path,
                  prefix: _Path,
                  file_op: Callable[[_Path, _Path], None],
                  excludes: Optional[list[str]] = None) -> None:
        file_list: list[str] = []
        base = self.directory
        for path, _, files in os.walk(os.path.abspath(root_dir)):
            for name in files:
                file_source = os.path.join(path, name)
                file_list_path = os.path.join(
                    prefix, os.path.relpath(file_source, root_dir))
                file_target = os.path.join(base, file_list_path)
                if excludes is None:
                    file_list.append(file_list_path)
                    file_op(file_source, file_target)
                else:
                    match_path = os.path.normpath(
                        os.path.join("/", file_list_path))
                    for exclude in excludes:
                        if fnmatch.fnmatch(match_path, exclude):
                            logging.info(
                                "%s: exclude file for pattern '%s': %s",
                                self.uid, exclude, file_target)
                            break
                    else:
                        file_list.append(file_list_path)
                        file_op(file_source, file_target)
        self.add_files(file_list)

    def add_tree(self,
                 root_dir: _Path,
                 prefix: _Path = ".",
                 excludes: Optional[list[str]] = None) -> None:
        """
        Add the files of the directory tree starting at the root directory
        to the file set of the directory state.

        The added file path is relative to the root directory.  The prefix is
        prepended to the file path for each file before it is added to the
        directory state.  The files are not copied or moved.
        """
        self._add_tree(root_dir, prefix, _file_nop, excludes)

    def copy_tree(self,
                  root_dir: _Path,
                  prefix: _Path = ".",
                  excludes: Optional[list[str]] = None) -> None:
        """
        Add the files of the directory tree starting at the root directory
        to the file set of the directory state.

        The added file path is relative to the root directory.  The prefix is
        prepended to the file path for each file before it is added to the
        directory state.  The files are copied.
        """
        self._add_tree(root_dir, prefix, self._copy_file, excludes)

    def move_tree(self,
                  root_dir: _Path,
                  prefix: _Path = ".",
                  excludes: Optional[list[str]] = None) -> None:
        """
        Add the files of the directory tree starting at the root directory
        to the file set of the directory state.

        The added file path is relative to the root directory.  The prefix is
        prepended to the file path for each file before it is added to the
        directory state.  The files are moved.
        """
        self._add_tree(root_dir, prefix, self._move_file, excludes)

    def add_tarfile_members(self, archive: _Path, prefix: _Path,
                            extract: bool) -> None:
        """
        Append the members of the archive to the file list of the directory
        state.

        For each member the prefix path and the member path are joined and then
        added to the file list of the directory state.  If extract is true,
        then the members of the archive are extracted to the prefix path.
        """
        extract_info = "and extract " if extract else ""
        logging.info("%s: add %smembers of '%s' using prefix '%s'", self.uid,
                     extract_info, archive, prefix)

        uid = os.getuid()
        gid = os.getgid()

        def member_filter(info: tarfile.TarInfo,
                          _path: str) -> tarfile.TarInfo:
            info.uid = uid
            info.gid = gid
            return info

        with tarfile.open(archive, "r") as tar_file:
            base = self.directory
            file_list = [
                os.path.relpath(os.path.join(prefix, info.name), base)
                for info in tar_file.getmembers() if not info.isdir()
            ]
            if extract:
                tar_file.extractall(prefix,
                                    numeric_owner=True,
                                    filter=member_filter)
            self.add_files(file_list)

    def create_symbolic_links(
            self, symbolic_link_list: list[dict[str, str]]) -> None:
        """
        Create the specified symbolic links in the directory state.
        """
        base = self.directory
        for symbolic_link in symbolic_link_list:
            link_name = symbolic_link["link-name"]
            target = symbolic_link["target"]
            link_path = os.path.join(base, link_name)
            target_path = os.path.join(base, target)
            target_path = os.path.relpath(target_path,
                                          os.path.dirname(link_path))
            logging.info("%s: create symbolic link from '%s' to '%s'",
                         self.uid, link_path, target_path)
            os.symlink(target_path, link_path)
            self.add_files([link_name])

    def _get_files_and_reverse_mapping(
        self, file_mappings: Optional[list[dict]]
    ) -> tuple[dict[str, str | None], dict[str, str]]:
        if not file_mappings:
            return self._files.copy(), {}
        mapped_files: dict[str, str | None] = {}
        reverse_mapping: dict[str, str] = {}
        mappings: list[tuple] = []
        for file_mapping in file_mappings:
            mappings.append(
                (re.compile(file_mapping["pattern"]),
                 file_mapping["replacement"], file_mapping.get("count", 0),
                 file_mapping.get("continue", False)))
        for file, digest in self._files.items():
            file_2 = file
            for index, mapping in enumerate(mappings):
                file_3 = file_2
                file_2, count = mapping[0].subn(mapping[1],
                                                file_2,
                                                count=mapping[2])
                if count:
                    logging.info("%s: mapping %s maps %s to %s", self.uid,
                                 index, file_3, file_2)
                    if not mapping[3]:
                        break
            mapped_files[file_2] = digest
            reverse_mapping[file_2] = file
        return mapped_files, reverse_mapping

    def lazy_clone(self,
                   other: "DirectoryStateBase",
                   file_mappings: Optional[list[dict]] = None) -> str:
        """ Lazily clone the directory state. """
        # pylint: disable=protected-access
        logging.info("%s: lazy clone from: %s", self.uid, other.uid)
        mapped_files, reverse_mapping = other._get_files_and_reverse_mapping(
            file_mappings)
        current = set(self._files.keys())
        new = set(mapped_files.keys())
        base = self.directory
        other_base = other.directory
        for file in sorted(current.difference(new)):
            target = os.path.join(base, file)
            try:
                logging.info("%s: remove: %s", self.uid, target)
                os.remove(target)
            except FileNotFoundError:
                logging.warning("%s: file not found: %s", self.uid, target)
        for file in sorted(new.difference(current)):
            source = reverse_mapping.get(file, file)
            target = os.path.join(base, file)
            self._copy_file(os.path.join(other_base, source), target)
        for file in sorted(current.intersection(new)):
            source = reverse_mapping.get(file, file)
            target = os.path.join(base, file)
            if self._files[file] == mapped_files[file]:
                logging.info("%s: keep as is: %s", self.uid, target)
            else:
                self._copy_file(os.path.join(other_base, source), target)
        self._files = mapped_files
        return self._add_hashes(base, self._get_hash)

    def json_dump(self, data: Any) -> None:
        """ Dump the data into the file of the directory state. """
        file_path = self.file
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "w", encoding="utf-8") as file:
            json.dump(data, file, sort_keys=True, indent=2)

    def json_load(self) -> Any:
        """ Load the data from the file of the directory state. """
        with open(self.file, "r", encoding="utf-8") as file:
            return json.load(file)

    def save(self) -> None:
        """ Save the directory state to the item file. """
        self.item.save()

    def needs_build(self) -> bool:
        return self.item["hash"] is None

    def is_present(self) -> bool:
        if not self._files:
            return False
        for digest in self._files.values():
            if digest is None:
                return False
        return True

    def discard(self) -> None:
        """ Discard the directory state. """
        logging.info("%s: discard", self.uid)
        self._discarded_files.update(self.remove_files())
        self.invalidate()
        self.save()

    def refresh(self) -> None:
        logging.info("%s: refresh", self.uid)
        self.load()
        count = len(self._files)
        if count == 1:
            self.description.add(
                f"Represent the file {self.description.path(self.file)}.")
        elif count < 14:
            self.description.add_list(
                [self.description.path(name) for name in self.files()],
                "Represent the files:")
        else:
            self.description.add(f"""Represent {count} files in directory
{self.description.path(self.directory)}.""")

    def commit(self, reason: str) -> None:
        files = self._discarded_files.union(self._files.keys())
        self._discarded_files.clear()
        if files:
            base = self.directory
            self.git_add([os.path.join(base, path) for path in sorted(files)])
        super().commit(reason)


def _make_digest(value: Optional[str]) -> str:
    assert value is not None
    return base64_to_hex_text(value)


def _gather_files_and_hashes(build_item: BuildItem, names: list[str],
                             base: Optional[str], relpath: str,
                             files_and_hashes: set[tuple[str, str]]) -> None:
    if not names:
        return
    for name in names[0].split(" "):
        for input_item in build_item.inputs(name):
            if isinstance(input_item, DirectoryStateBase):
                files_and_hashes.update(
                    (os.path.relpath(file, relpath), _make_digest(digest))
                    for file, digest in input_item.files_and_hashes(base))
            _gather_files_and_hashes(input_item, names[1:], base, relpath,
                                     files_and_hashes)


class DirectoryStateValueProvider(ItemValueProvider):
    """ Provides values related to directory states. """

    # pylint: disable=too-few-public-methods

    def __init__(self, build_item: BuildItem) -> None:
        mapper = build_item.mapper
        super().__init__(mapper)
        self._build_item = build_item
        type_name = build_item.item.type
        mapper.add_get_value(f"{type_name}:/input-file-list",
                             self._get_input_files)
        mapper.add_get_value(f"{type_name}:/input-file-list-with-hashes",
                             self._get_input_files_and_hashes)

    def _files_and_hashes(self,
                          ctx: ItemGetValueContext) -> list[tuple[str, str]]:
        files_and_hashes: set[tuple[str, str]] = set()
        args, kwargs = ctx.unpack_args_dict(ctx.mapper.substitute)
        base = kwargs.get("base")
        relpath = kwargs.get("relpath", "/")
        _gather_files_and_hashes(self._build_item, args, base, relpath,
                                 files_and_hashes)
        return sorted(files_and_hashes)

    def _get_input_files(self, ctx: ItemGetValueContext) -> str:
        mapper = self.mapper
        assert isinstance(mapper, BuildItemMapper)
        content = mapper.create_content()
        content.add_list(f"{content.path(file)}"
                         for file, _ in self._files_and_hashes(ctx))
        return content.join()

    def _get_input_files_and_hashes(self, ctx: ItemGetValueContext) -> str:
        mapper = self.mapper
        assert isinstance(mapper, BuildItemMapper)
        content = mapper.create_content()
        content.add_list(
            f"{content.path(file)} with an SHA512 digest of {digest}"
            for file, digest in self._files_and_hashes(ctx))
        return content.join()



# SPDX-License-Identifier: BSD-2-Clause
""" Builds an Software Verification Report (SVR). """

# Copyright (C) 2022, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import itertools
from typing import Iterator

from specitems import COL_SPAN, Item, ItemGetValueContext, SphinxContent
from specware import gather_benchmarks_and_test_suites

from .directorystate import DirectoryState
from .membench import generate
from .pkgitems import PackageBuildDirector
from .linkhub import LinkHub
from .sphinxbuilder import spacify
from .specdocbuilder import SpecDocumentBuilder
from .testaggregator import get_test_result_status, TestAggregator

_DEFINE_NOT_DEFINED = "requirement/functional/interface-define-not-defined"

_GROUP = ["interface/group", "requirement/non-functional/design-group"]


def _gather_design_components(item: Item, components: list[Item]) -> None:
    if item.type in _GROUP:
        components.append(item)
        return
    for parent in item.parents("interface-function"):
        components.append(parent)
    for parent in item.parents("requirement-refinement"):
        if parent.type in _GROUP:
            components.append(parent)
    if not components:
        for parent in item.parents("requirement-refinement"):
            _gather_design_components(parent, components)


def _get_name(kind_name: str) -> str:
    return kind_name[kind_name.find("/") + 1:]


def _get_kind_name(kind_name: str) -> tuple[str, str]:
    separator = kind_name.find("/")
    return kind_name[0:separator], kind_name[separator + 1:]


class SVRBuilder(SpecDocumentBuilder):
    """ Builds an Software Verification Report (SVR). """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        my_type = self.item.type
        self.mapper.add_get_value(
            f"{my_type}:/traceability-requirements-to-design",
            self._traceability_requirements_to_design)
        self.mapper.add_get_value(
            f"{my_type}:/traceability-design-to-requirements",
            self._traceability_design_to_requirements)
        self.mapper.add_get_value(f"{my_type}:/traceability-design-to-code",
                                  self._traceability_design_to_code)
        self.mapper.add_get_value(f"{my_type}:/traceability-code-to-design",
                                  self._traceability_code_to_design)
        self.mapper.add_get_value(f"{my_type}:/unit-verification",
                                  self._unit_verification)
        self.mapper.add_get_value(f"{my_type}:/code-coverage-achievement",
                                  self._code_coverage_achievement)
        self.mapper.add_get_value(f"{my_type}:/performance-summary",
                                  self._performance_summary)
        self.mapper.add_get_value(f"{my_type}:/memory-benchmarks",
                                  self._get_membench)
        self._design_to_code: list[tuple[str, str]] = []

    def _req_to_design(self) -> Iterator[tuple[Item, list[Item]]]:
        for item in self.spec.get_related_requirements():
            components: list[Item] = []
            _gather_design_components(item, components)
            yield (item, components)

    def _traceability_requirements_to_design(self,
                                             _ctx: ItemGetValueContext) -> str:
        content = SphinxContent(section_level=2)
        rows: list[tuple[str | int,
                         ...]] = [("Requirement", "Design Component")]
        for item, components in self._req_to_design():
            req: str | int = self.mapper.get_link(item)
            for component in sorted(set(components)):
                try:
                    link = component.view["document-links"]["sdd"]
                except KeyError:
                    group = "requirement/non-functional/design-group"
                    if item.type == _DEFINE_NOT_DEFINED:
                        link = "N/A (interface define is not defined)"
                    elif item.type == group and item["identifier"] is None:
                        link = "N/A (external design)"
                    else:
                        link = "**no reference to SDD**"
                rows.append((req, link))
                req = COL_SPAN
            if isinstance(req, str) and req:
                rows.append(
                    (req, "N/A (no directly associated design components)"))
        content.add_grid_table(rows, [50, 50], font_size=-4)
        return str(content)

    def _traceability_design_to_requirements(self,
                                             _ctx: ItemGetValueContext) -> str:
        content = SphinxContent(section_level=2)
        link_hub = self.input("link-hub")
        assert isinstance(link_hub, LinkHub)
        rows = [("Design Component", "Requirement")]
        for _, info in sorted(link_hub.name_info.items(),
                              key=lambda x: _get_name(x[0])):
            if not isinstance(info, dict):
                continue
            variants = info.get("variants", [])
            for variant in itertools.chain([info], variants):
                name = f"`{spacify(variant['name'])} <{variant['link']}>`__"
                if variants:
                    file_path = variant["file"]
                    file_link = link_hub.name_info[f"file/{file_path}"]["link"]
                    name = f"{name} in `{spacify(file_path)} <{file_link}>`__"
                for item in sorted(variant["items"]):
                    url = item.view["document-links"]["url"]
                    rows.append((name, f"`{item.spec_2} <{url}>`__"))
        content.add_grid_table(rows, [60, 40], font_size=-4)
        return str(content)

    def _get_design_to_code(self) -> list[tuple[str, str]]:
        if self._design_to_code:
            return self._design_to_code
        link_hub = self.input("link-hub")
        assert isinstance(link_hub, LinkHub)
        design_to_code: list[tuple[str, str]] = []
        for kind_name, info in sorted(link_hub.name_info.items(),
                                      key=lambda x: _get_name(x[0])):
            if not isinstance(info, dict):
                continue
            kind, name = _get_kind_name(kind_name)
            if kind == "file":
                continue
            if kind == "group":
                name = f"`{spacify(info['name'])} <{info['link']}>`__"
                for file_path in info["files"]:
                    file_link = link_hub.name_info[f"file/{file_path}"]["link"]
                    file = f"`{spacify(file_path)} <{file_link}>`__"
                    design_to_code.append((name, file))
                continue
            variants = info.get("variants", [])
            for variant in itertools.chain([info], variants):
                name = f"`{spacify(variant['name'])} <{variant['link']}>`__"
                file_path = variant["file"]
                file_link = link_hub.name_info[f"file/{file_path}"]["link"]
                file = f"`{spacify(file_path)} <{file_link}>`__"
                design_to_code.append((name, file))
        self._design_to_code = design_to_code
        return design_to_code

    def _traceability_design_to_code(self, _ctx: ItemGetValueContext) -> str:
        content = SphinxContent(section_level=2)
        rows: list[tuple[str | int, ...]] = [("Design Component", "File")]
        last = ""
        for design, code in self._get_design_to_code():
            rows.append((COL_SPAN if last == design else design, code))
            last = design
        content.add_grid_table(rows, [60, 40], font_size=-4)
        return str(content)

    def _traceability_code_to_design(self, _ctx: ItemGetValueContext) -> str:
        content = SphinxContent(section_level=2)
        code_to_design = sorted(
            (code, design) for design, code in self._get_design_to_code())
        rows: list[tuple[str | int, ...]] = [("File", "Design Component")]
        last: str | int = COL_SPAN
        for code, design in code_to_design:
            rows.append((COL_SPAN if last == code else code, design))
            last = code
        content.add_grid_table(rows, [40, 60], font_size=-4)
        return str(content)

    def _unit_verification(self, _ctx: ItemGetValueContext) -> str:
        content = SphinxContent(section_level=2)
        test_suites: list[Item] = []
        gather_benchmarks_and_test_suites(self.item.cache["/testsuites/unit"],
                                          test_suites)
        rows = [("Test Case", "Status")]
        unspecified: dict[str, list[str]] = {}
        for test_suite in sorted(test_suites):
            for test_case in test_suite.children("test-case"):
                rows.append((self.mapper.get_link(test_case),
                             get_test_result_status(test_case, "", "")))
            for test_results in test_suite.view.get("test-results",
                                                    {}).values():
                for data in test_results:
                    for test_case in data["unspecified-test-cases"]:
                        unspecified.setdefault(test_case["name"], []).append(
                            f"`{data['status']} <{data['link']}>`__")
        rows_2 = [["Test Case without Specification", "Status"]]
        for name, status in sorted(unspecified.items()):
            rows_2.append([name, ", ".join(status)])
        content.add_grid_table(rows, [80, 20], font_size=-1)
        content.add_grid_table(rows_2, [80, 20], font_size=-1)
        return str(content)

    def _code_coverage_achievement(self, _ctx: ItemGetValueContext) -> str:
        content = SphinxContent(section_level=2)
        test_aggregator = self.input("test-aggregation")
        assert isinstance(test_aggregator, TestAggregator)
        test_aggregator.add_coverage_achievement(content)
        return str(content)

    def _performance_summary(self, _ctx: ItemGetValueContext) -> str:
        content = SphinxContent(section_level=2)
        types = ("requirement/non-functional/performance-runtime", )
        rows = [("Requirement", "Status")]
        for item in self.spec.get_related_items_by_type(types):
            rows.append((self.mapper.get_link(item),
                         get_test_result_status(item, "", "")))
        content.add_grid_table(rows, [80, 20], font_size=-1)
        return str(content)

    def _get_membench(self, ctx: ItemGetValueContext) -> str:
        with self.section_level_scope(ctx):
            content = SphinxContent(section_level=self.section_level)
            link, membench = self.input_link("membench-results")
            assert isinstance(membench, DirectoryState)
            label = membench.substitute(link["build-label"])
            sections_by_uid = membench.json_load()[label]["membench"]
            root = self.item.cache["/rtems/req/mem-basic"]
            table_pivots = ["/rtems/req/mem-smp-1"]
            generate(content, sections_by_uid, root, table_pivots, self.mapper)
            return "\n".join(content)



# SPDX-License-Identifier: BSD-2-Clause
""" Produces a directory state which is subset of a repository. """

# Copyright (C) 2021 EDISOFT
# Copyright (C) 2020, 2025 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging
import os
import tempfile

from specitems import ItemCache, ItemCacheConfig, is_enabled
from specware import gather_build_files

from .directorystate import DirectoryState


class RepositorySubset(DirectoryState):
    """
    Creates a directory containing only the specified subset of the
    repository.
    """

    def gather_files(self) -> list[str]:
        """ Gather the files of the repository subset. """
        logging.info("%s: gather repository subset", self.uid)
        enabled_set = list(self.enabled_set)
        for enable_enabled_by in self["enabled-set-extension"]:
            if is_enabled(enabled_set, enable_enabled_by["enabled-by"]):
                enabled_set.append(enable_enabled_by["enable"])
        build_spec_directories = self["build-spec-directories"]
        if build_spec_directories:
            with tempfile.TemporaryDirectory() as tmp_dir:
                cache_config = ItemCacheConfig(
                    paths=[
                        os.path.join(self.directory, directory)
                        for directory in build_spec_directories
                    ],
                    resolve_proxies=False,
                    cache_directory=tmp_dir,
                )
                item_cache = ItemCache(cache_config)
                logging.info("items %s %s", cache_config,
                             list(item_cache.keys()))
        else:
            item_cache = self.item.cache
        config = {
            "arch": self["arch"],
            "bsp": self["bsp"],
            "enabled-set": enabled_set,
            "base-directory-map": self["base-directory-map"],
            "build-uids": self["build-uids"]
        }
        files = gather_build_files(config, item_cache,
                                   self["include-test-header"])
        files.extend(self["extra-files"])
        return files

    def run(self):
        source = self.input("source")
        assert isinstance(source, DirectoryState)

        self.discard()
        self.clear()

        logging.info("%s: copy gathered files", self.uid)
        self.copy_files(source.directory, self.gather_files())

        self.description.add(f"""Produce a repository subset in directory
{self.description.path(self.directory)}""")



# SPDX-License-Identifier: BSD-2-Clause
""" Provides a command line interface to make items of a package. """

# Copyright (C) 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import argparse
import logging
import sys

from specitems import (ItemCache, ItemCacheConfig, get_item_cache_arguments,
                       verify_specification_format)

from .pkgfactory import create_build_item_factory
from .pkgitems import BuildItemTypeProvider, PackageBuildDirector


def _get_arguments(argv: list[str]) -> argparse.Namespace:

    def _add_arguments(parser):
        parser.add_argument(
            "--force",
            type=str,
            action="append",
            default=None,
            help="building items with an UID matching this pattern is forced")
        parser.add_argument(
            "--package-uid",
            help="the package component UID (default: /pkg/component)",
            default="/pkg/component")
        parser.add_argument("--spec-verify",
                            action="store_true",
                            help="verify the specification")
        parser.add_argument("--use-git",
                            action="store_true",
                            help="use Git to track build steps")
        parser.add_argument("uids",
                            metavar="UID",
                            nargs="*",
                            help="the UID of an item to make")

    return get_item_cache_arguments(argv,
                                    description=climake.__doc__,
                                    add_arguments=(_add_arguments, ))


def climake(argv: list[str] = sys.argv) -> None:
    """ Make items of a package. """

    args = _get_arguments(argv[1:])
    cache_config = ItemCacheConfig(paths=args.spec_directories,
                                   cache_directory=args.cache_directory)
    type_provider = BuildItemTypeProvider({})
    uid = args.package_uid
    logging.info("%s: load item cache", uid)
    item_cache = ItemCache(cache_config, type_provider=type_provider)
    logging.info("%s: loading done", uid)
    if args.spec_verify:
        logger = logging.getLogger()
        level = logger.getEffectiveLevel()
        logger.setLevel(logging.ERROR)
        status = verify_specification_format(item_cache)
        logger.setLevel(level)
        if status.critical or status.error:
            return
    factory = create_build_item_factory()
    director = PackageBuildDirector(item_cache,
                                    uid,
                                    factory,
                                    use_git=args.use_git)
    director.build_only(args.uids, force_patterns=args.force)



# SPDX-License-Identifier: BSD-2-Clause
""" Provides utility functions. """

# Copyright (C) 2020, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import argparse
import datetime
import itertools
import logging
import os
import shutil
from typing import Callable, Iterable, Optional

from specitems import ItemMapper, get_arguments


def now_utc() -> str:
    """ Return the current UTC time in ISO 8601 format. """
    return datetime.datetime.now(datetime.timezone.utc).isoformat()


def duration(value: float | str) -> str:
    """ Convert a duration in seconds into a value with unit string. """
    if isinstance(value, str):
        return value
    assert value >= 0.0
    if value == 0.0:
        return "0s"
    if value < 1e-6:
        return f"{value * 1e9:.3f}ns"
    if value < 1e-3:
        return f"{value * 1e6:.3f}μs"
    if value < 1.0:
        return f"{value * 1e3:.3f}ms"
    return f"{value:.3f}s"


def copy_file(src_file: str, dst_file: str, log_context: str) -> None:
    """ Copy the source file to the destination file. """
    os.makedirs(os.path.dirname(dst_file), exist_ok=True)
    logging.info("%s: copy '%s' to '%s'", log_context, src_file, dst_file)
    shutil.copy2(src_file, dst_file)


def copy_files(src_dir: str, dst_dir: str, files: list[str],
               log_context: str) -> None:
    """
    Copy a list of files in the source directory to the destination directory
    preserving the directory of the files relative to the source directory.
    """
    for a_file in files:
        src_file = os.path.join(src_dir, a_file)
        dst_file = os.path.join(dst_dir, a_file)
        os.makedirs(os.path.dirname(dst_file), exist_ok=True)
        logging.info("%s: copy '%s' to '%s'", log_context, src_file, dst_file)
        shutil.copy2(src_file, dst_file)


def copy_and_substitute(src_file: str, dst_file: str, mapper: ItemMapper,
                        log_context: str) -> None:
    """
    Copy the file from the source to the destination path and performs a
    variable substitution on the file content using the item mapper.
    """
    logging.info("%s: read: %s", log_context, src_file)
    with open(src_file, "r", encoding="utf-8") as src:
        logging.info("%s: substitute using mapper of item %s", log_context,
                     mapper.item.uid)
        content = mapper.substitute(src.read())
        logging.info("%s: write: %s", log_context, dst_file)
        os.makedirs(os.path.dirname(dst_file), exist_ok=True)
        with open(dst_file, "w+", encoding="utf-8") as dst:
            dst.write(content)


def remove_empty_directories(scope: str, base: str) -> None:
    """
    Recursively remove all empty subdirectories of base and base itself if it
    gets empty.

    The scope is used for log messages.
    """
    removed: set[str] = set()
    for root, subdirs, files in os.walk(base, topdown=False):
        if files:
            continue
        if any(subdir for subdir in subdirs
               if os.path.join(root, subdir) not in removed):
            continue
        logging.info("%s: remove empty directory: %s", scope, root)
        os.rmdir(root)
        removed.add(root)


def get_build_arguments(
    argv: list[str],
    default_log_level: str = "INFO",
    description: Optional[str] = None,
    add_arguments: Iterable[Callable[[argparse.ArgumentParser],
                                     None]] = tuple(),
    post_process_arguments: Iterable[Callable[[argparse.Namespace],
                                              None]] = tuple()
) -> argparse.Namespace:
    """
    Create an argument parser with default logging and build options,
    optionally add arguments to the parser, parse the argument vector,
    initialize logging, optionally post process the parsed arguments, and
    return the parsed arguments.
    """

    def _add_arguments(parser):
        parser.add_argument("--only",
                            type=str,
                            action="append",
                            default=None,
                            help="build only these steps")
        parser.add_argument("--force",
                            type=str,
                            action="append",
                            default=None,
                            help="force to build these steps")
        parser.add_argument("--skip",
                            type=str,
                            action="append",
                            default=None,
                            help="skip these steps")
        parser.add_argument("--no-spec-verify",
                            action="store_true",
                            help="do not verify the specification")

    return get_arguments(argv, default_log_level, description,
                         itertools.chain((_add_arguments, ), add_arguments),
                         post_process_arguments)



# SPDX-License-Identifier: BSD-2-Clause
""" Provides links from items to documents. """

# Copyright (C) 2022, 2025 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import functools
import itertools
import os
from typing import Any, Dict, List, Optional
from xml.etree import ElementTree

from specitems import Item, ItemGetValueContext, make_label

from .directorystate import DirectoryState
from .pkgitems import BuildItem, BuildItemMapper, PackageBuildDirector
from .rtems import RTEMSItemCache

_EB_PKG_SUPPORT = "https://embedded-brains.de/qdp-support"

# Sorted by expected frequency
_ELEMENT_KINDS = ("function", "define", "group", "file", "type", "enum",
                  "object", "enumerator")


def _member_link(elem: Any) -> str:
    return f"{elem.find('anchorfile').text}#{elem.find('anchor').text}"


def _arglist(elem: Any) -> str:
    arglist = elem.find('arglist').text
    if not arglist:
        return ""
    return arglist


def _type(elem: Any) -> str:
    return elem.find('type').text


def _sdd_link(name_info: Dict[str, Any], link: str) -> str:
    return f"{name_info['dir/sdd']}/{link}"


def _get_file_path(elem: Any) -> str:
    path = elem.find("path").text
    if not path:
        path = ""
    return f"{path}{elem.find('name').text}"


def _add_doxygen_item(name_info: Dict[str, Any], elem: Any, parent: Any,
                      name: str, elem_info: Dict[str, Any]) -> None:
    parent_kind = parent.attrib.get("kind", "")
    if parent_kind == "class":
        return
    if parent_kind == "file":
        elem_info["file"] = _get_file_path(parent)
    anchorfile = elem.find("anchorfile")
    if anchorfile is not None:
        elem_info_2 = name_info.setdefault(name, {}).setdefault(
            "instances", {}).setdefault(anchorfile.text, {})
    else:
        elem_info_2 = name_info.setdefault(name, {})
    elem_info_2.update(elem_info)
    if parent_kind == "group":
        elem_info_2.setdefault("groups", set()).add(parent.find("name").text)


def _doxygen_define(name_info: Dict[str, Any], elem: Any, parent: Any) -> None:
    name = elem.find("name").text
    arglist = _arglist(elem)
    elem_info = {
        "doxygen": f"@def {name}{arglist}",
        "link": _sdd_link(name_info, _member_link(elem)),
        "name": f"{name}()" if arglist else name
    }
    _add_doxygen_item(name_info, elem, parent, f"define/{name}", elem_info)


def _doxygen_enumeration(name_info: Dict[str, Any], elem: Any,
                         parent: Any) -> None:
    name = elem.find("name").text
    elem_info = {
        "doxygen": f"@enum {name}",
        "link": _sdd_link(name_info, _member_link(elem)),
        "name": name
    }
    _add_doxygen_item(name_info, elem, parent, f"enum/{name}", elem_info)


def _doxygen_enumvalue(name_info: Dict[str, Any], elem: Any,
                       parent: Any) -> None:
    name = elem.find("name").text
    elem_info = {
        "doxygen": f"@var {name}",
        "link": _sdd_link(name_info, _member_link(elem)),
        "name": name
    }
    _add_doxygen_item(name_info, elem, parent, f"enumerator/{name}", elem_info)


def _doxygen_file(name_info: Dict[str, Any], elem: Any, parent: Any) -> None:
    prefix = f"{name_info['dir/source']}"
    name = _get_file_path(elem)
    name_info[f"file-key/{name}"] = name
    elem_info = {
        "doxygen": f"@file {name}",
        "link": _sdd_link(name_info,
                          elem.find('filename').text),
        "name": name,
        "prefix": prefix
    }
    _add_doxygen_item(name_info, elem, parent, f"file/{name}", elem_info)


# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
def _doxygen_decl(name_info: Dict[str, Any],
                  elem: Any,
                  parent: Any,
                  kind: str,
                  cmd: str,
                  postfix: str = "") -> None:
    name = elem.find("name").text
    elem_info = {
        "doxygen": f"@{cmd} {_type(elem)} {name}{_arglist(elem)}",
        "link": _sdd_link(name_info, _member_link(elem)),
        "name": f"{name}{postfix}"
    }
    _add_doxygen_item(name_info, elem, parent, f"{kind}/{name}", elem_info)


def _doxygen_function(name_info: Dict[str, Any], elem: Any,
                      parent: Any) -> None:
    _doxygen_decl(name_info, elem, parent, "function", "fn", "()")


def _doxygen_group(name_info: Dict[str, Any], elem: Any, parent: Any) -> None:
    file_keys: List[str] = []
    for key in elem.findall("file"):
        file_keys.append(f"{key.attrib.get('path', '')}{key.text}")
    subgroups: List[str] = []
    for key in elem.findall("subgroup"):
        subgroups.append(key.text)
    group = elem.find("name").text
    name_info.setdefault("groups/all", []).append(group)
    elem_info = {
        "doxygen": f"@addtogroup {group}",
        "link": _sdd_link(name_info,
                          elem.find("filename").text),
        "name": elem.find("title").text,
        "file-keys": file_keys,
        "subgroups": subgroups
    }
    _add_doxygen_item(name_info, elem, parent, f"group/{group}", elem_info)


def _doxygen_type(name_info: Dict[str, Any], elem: Any, parent: Any,
                  kind: str) -> None:
    if elem.tag == "compound":
        name = elem.find("name").text
        elem_info = {
            "doxygen": f"@{kind} {name}",
            "link": _sdd_link(name_info,
                              elem.find("filename").text)
        }
    else:
        # This is required to get the file of compounds
        assert elem.tag == "class"
        name = elem.text
        elem_info = {}
    elem_info["name"] = name
    _add_doxygen_item(name_info, elem, parent, f"type/{name}", elem_info)


def _doxygen_struct(name_info: Dict[str, Any], elem: Any, parent: Any) -> None:
    _doxygen_type(name_info, elem, parent, "struct")


def _doxygen_typedef(name_info: Dict[str, Any], elem: Any,
                     parent: Any) -> None:
    _doxygen_decl(name_info, elem, parent, "type", "typedef")


def _doxygen_union(name_info: Dict[str, Any], elem: Any, parent: Any) -> None:
    _doxygen_type(name_info, elem, parent, "union")


def _doxygen_variable(name_info: Dict[str, Any], elem: Any,
                      parent: Any) -> None:
    if parent.attrib.get("kind", "") in ("struct", "union"):
        return
    if "::" in elem.find("name").text:
        # Compound members may show up in groups.  In this case, they have a
        # "::" in the name.  Do not add a name information for them.
        return
    _doxygen_decl(name_info, elem, parent, "object", "var")


def _doxygen_nop(_name_info: Dict[str, Any], _elem: Any, _parent: Any) -> None:
    pass


_DOXYGEN_KINDS = {
    "define": _doxygen_define,
    "enumeration": _doxygen_enumeration,
    "enumvalue": _doxygen_enumvalue,
    "file": _doxygen_file,
    "function": _doxygen_function,
    "group": _doxygen_group,
    "struct": _doxygen_struct,
    "typedef": _doxygen_typedef,
    "union": _doxygen_union,
    "variable": _doxygen_variable
}


def _depth_iter(element):
    stack = [(iter([element]), None)]
    while stack:
        element_2 = next(stack[-1][0], None)
        if element_2 is None:
            stack.pop()
        else:
            stack.append((iter(element_2), element_2))
            yield (element_2, stack[-2][1])


def _gather_name_info(name_info: Dict[str, Any], tagfile: str) -> None:
    tree = ElementTree.parse(tagfile)
    for elem, parent in _depth_iter(tree.getroot()):
        try:
            kind = elem.attrib["kind"]
        except KeyError:
            continue
        _DOXYGEN_KINDS.get(kind, _doxygen_nop)(name_info, elem, parent)


def _name_info_key_default(item: Item) -> str:
    raise ValueError(f"no name information for: {item.uid}")


def _name_info_key_define(item: Item) -> str:
    return f"define/{item['name']}"


def _name_info_key_enum(item: Item) -> str:
    return f"enum/{item['name']}"


def _name_info_key_enumerator(item: Item) -> str:
    return f"enumerator/{item['name']}"


def _name_info_key_header_file(item: Item) -> str:
    return f"file/{item['prefix']}/{item['path']}"


def _name_info_key_forward_decl(item: Item) -> str:
    target = item.parent("interface-target")
    return f"{target['interface-type']}/{target['name']}"


def _name_info_key_function(item: Item) -> str:
    return f"function/{item['name']}"


def _name_info_key_group(item: Item) -> str:
    return f"group/{item['identifier']}"


def _name_info_key_appl_config_group(item: Item) -> str:
    return f"group/RTEMSApplConfig{item['name'].replace(' ', '')}"


def _name_info_key_object(item: Item) -> str:
    return f"object/{item['name']}"


def _name_info_key_type(item: Item) -> str:
    return f"type/{item['name']}"


def _link_default(item: Item, _link: str) -> str:
    raise ValueError(f"no link for: {item.uid}")


def _link_design_group(item: Item, link: str) -> str:
    return f"`{item.view['sdd-info']['name']} <{link}>`__"


def _link_name(item: Item, link: str) -> str:
    return f"`{item['name']} <{link}>`__"


def _link_forward_decl(item: Item, link: str) -> str:
    return _link_name(item.parent("interface-target"), link)


def _link_function(item: Item, link: str) -> str:
    return f"`{item['name']}() <{link}>`__"


def _link_header_file(item: Item, link: str) -> str:
    return f"`<{item['path']}> <{link}>`__"


def _link_unspecified_type(item: Item, link: str) -> str:
    return f"`{item['interface-type'][12:]} {item['name']} <{link}>`__"


def spec_label(item: Item) -> str:
    """ Returns the specification label of the item. """
    return make_label(item.spec)


def _ref_default(item: Item) -> str:
    raise ValueError(f"no reference for: {item.uid}")


def _ref_name(item: Item) -> str:
    return f":ref:`{item['name']} <{spec_label(item)}>`"


def _ref_forward_decl(item: Item) -> str:
    return _ref_name(item.parent("interface-target"))


def _ref_function(item: Item) -> str:
    return f":ref:`{item['name']}() <{spec_label(item)}>`"


def _ref_header_file(item: Item) -> str:
    return f":ref:`<{item['path']}> <{spec_label(item)}>`"


def _ref_unspecified_type(item: Item) -> str:
    return (f":ref:`{item['interface-type'][12:]} "
            f"{item['name']} <{spec_label(item)}>`")


def _ref_spec(item: Item) -> str:
    return f":ref:`{item.spec_2} <{spec_label(item)}>`"


def _ref_disabled_name(item: Item) -> str:
    return f"``{item['name']}``"


def _ref_disabled_forward_decl(item: Item) -> str:
    return _ref_disabled_name(item.parent("interface-target"))


def _ref_disabled_function(item: Item) -> str:
    return f"``{item['name']}()``"


def _ref_disabled_header_file(item: Item) -> str:
    return f"``<{item['path']}>``"


def _ref_disabled_spec(item: Item) -> str:
    return item.spec_2


def _spec_link(item: Item, link: str) -> str:
    return f"`{item.spec_2} <{link}>`__"


def _augment_default(item: Item, _name_info: Dict[str, Any]) -> None:
    item.view["default-document"] = "none"
    item.view["document-links"] = {"none": f"``{item.spec_2}``"}


def _add_doxygen_info(item: Item, name_info: Dict[str, Any]) -> None:
    try:
        elem_info = name_info[_ITEM_SPECIFICS[item.type][5](item)]
    except KeyError:
        return
    item.view["sdd-info"] = elem_info
    path = elem_info.get("file", None)
    if path is not None:
        item.view["source-file"] = path
    get_link = _ITEM_SPECIFICS[item.type][4]
    item.view["document-links"]["sdd"] = get_link(item, f"{elem_info['link']}")


def _anchor(item: Item) -> str:
    return f"#spec{item.ident.lower()}"


def _augment_interface(item: Item, name_info: Dict[str, Any]) -> None:
    item.view["default-document"] = "icd"
    link = (f"{name_info['dir/icd']}/"
            f"requirements-and-design.html{_anchor(item)}")
    get_link = _ITEM_SPECIFICS[item.type][4]
    item.view["document-links"] = {
        "default": _spec_link(item, link),
        "url": link,
        "icd": get_link(item, link),
        "self": _ITEM_SPECIFICS[item.type][2](item),
    }
    _add_doxygen_info(item, name_info)


def _augment_requirement(item: Item, name_info: Dict[str, Any]) -> None:
    item.view["default-document"] = "srs"
    link = f"{name_info['dir/srs']}/requirements.html{_anchor(item)}"
    spec_link = _spec_link(item, link)
    item.view["document-links"] = {
        "default": spec_link,
        "url": link,
        "srs": spec_link,
        "self": _ITEM_SPECIFICS[item.type][2](item),
    }


def _augment_design_group(item: Item, name_info: Dict[str, Any]) -> None:
    _augment_requirement(item, name_info)
    _add_doxygen_info(item, name_info)


def _augment_requirement_self_test(item: Item, name_info: Dict[str,
                                                               Any]) -> None:
    _augment_requirement(item, name_info)
    item.view["document-links"]["test-plan"] = _spec_link(
        item, f"{name_info['dir/svs']}/"
        f"test-case-specification.html{_anchor(item)}")


def _augment_interface_requirement(item: Item, name_info: Dict[str,
                                                               Any]) -> None:
    item.view["default-document"] = "icd"
    link = (f"{name_info['dir/icd']}/"
            f"requirements-and-design.html{_anchor(item)}")
    spec_link = _spec_link(item, link)
    item.view["document-links"] = {
        "default": spec_link,
        "url": link,
        "icd": spec_link,
        "self": _ITEM_SPECIFICS[item.type][2](item),
    }


def _augment_test_case(item: Item, name_info: Dict[str, Any]) -> None:
    try:
        if "validation" in item.parent("test-case").parent(
                "requirement-refinement").uid:
            key = "svs"
        else:
            key = "suitp"
    except IndexError as err:
        raise ValueError("no test-case to test-suite to "
                         "requirement-refinement link for (likely cause is "
                         f"that no build item is enabled): {item.uid} -> "
                         f"{[item.uid for item in item.parents()]}") from err
    item.view["default-document"] = key
    prefix = name_info[f"dir/{key}"]
    link = f"{prefix}/test-case-specification.html{_anchor(item)}"
    spec_link = _spec_link(item, link)
    item.view["document-links"] = {
        "default": spec_link,
        "url": link,
        key: spec_link,
        "test-plan": spec_link,
        "self": _ITEM_SPECIFICS[item.type][2](item),
    }


def _augment_test_suite_for_document(item: Item, name_info: Dict[str, Any],
                                     document: str) -> None:
    item.view["default-document"] = document
    prefix = name_info[f"dir/{document}"]
    link = f"{prefix}/test-design.html{_anchor(item)}"
    spec_link = _spec_link(item, link)
    item.view["document-links"] = {
        "default": spec_link,
        "url": link,
        document: spec_link,
        "test-plan": spec_link,
        "self": _ITEM_SPECIFICS[item.type][2](item),
    }


def _augment_test_suite(item: Item, name_info: Dict[str, Any]) -> None:
    if "validation" in item.parent("requirement-refinement").uid:
        document = "svs"
    else:
        document = "suitp"
    _augment_test_suite_for_document(item, name_info, document)


def _augment_memory_benchmark(item: Item, name_info: Dict[str, Any]) -> None:
    _augment_test_suite_for_document(item, name_info, "svs")


def _augment_other_validation(item: Item, name_info: Dict[str, Any]) -> None:
    item.view["default-document"] = "svs"
    prefix = name_info["dir/svs"]
    link = f"{prefix}/validation-other.html{_anchor(item)}"
    spec_link = _spec_link(item, link)
    item.view["document-links"] = {
        "default": spec_link,
        "url": link,
        "svs": spec_link,
        "test-plan": spec_link,
        "self": _ITEM_SPECIFICS[item.type][2](item),
    }


_ITEM_DEFAULT = ("unknown", _augment_default, _ref_default, _ref_disabled_spec,
                 _link_default, _name_info_key_default)

_ITEM_SPECIFICS = {
    "constraint": ("constraint", _augment_requirement, _ref_spec,
                   _ref_disabled_spec, _link_default, _name_info_key_default),
    "glossary/group": ("glossary group", _augment_default, _ref_default,
                       _ref_disabled_spec, _link_default,
                       _name_info_key_default),
    "glossary/term": ("glossary term", _augment_default, _ref_default,
                      _ref_disabled_spec, _link_default,
                      _name_info_key_default),
    "interface/appl-config-group":
    ("application configuration group", _augment_interface, _ref_name,
     _ref_disabled_name, _link_name, _name_info_key_appl_config_group),
    "interface/appl-config-option/feature":
    ("application configuration option", _augment_interface, _ref_name,
     _ref_disabled_name, _link_name, _name_info_key_define),
    "interface/appl-config-option/feature-enable":
    ("application configuration option", _augment_interface, _ref_name,
     _ref_disabled_name, _link_name, _name_info_key_define),
    "interface/appl-config-option/initializer":
    ("application configuration option", _augment_interface, _ref_name,
     _ref_disabled_name, _link_name, _name_info_key_define),
    "interface/appl-config-option/integer":
    ("application configuration option", _augment_interface, _ref_name,
     _ref_disabled_name, _link_name, _name_info_key_define),
    "interface/define": ("define", _augment_interface, _ref_name,
                         _ref_disabled_name, _link_name,
                         _name_info_key_define),
    "interface/domain": ("interface domain", _augment_interface_requirement,
                         _ref_name, _ref_disabled_name, _link_name,
                         _name_info_key_group),
    "interface/enum": ("enumeration", _augment_interface, _ref_name,
                       _ref_disabled_name, _link_name, _name_info_key_enum),
    "interface/enumerator": ("enumerator", _augment_interface, _ref_name,
                             _ref_disabled_name, _link_name,
                             _name_info_key_enumerator),
    "interface/forward-declaration":
    ("forward declaration", _augment_interface, _ref_forward_decl,
     _ref_disabled_forward_decl, _link_forward_decl,
     _name_info_key_forward_decl),
    "interface/function": ("directive", _augment_interface, _ref_function,
                           _ref_disabled_function, _link_function,
                           _name_info_key_function),
    "interface/group": ("interface group", _augment_interface, _ref_name,
                        _ref_disabled_name, _link_name, _name_info_key_group),
    "interface/header-file": ("header file", _augment_interface,
                              _ref_header_file, _ref_disabled_header_file,
                              _link_header_file, _name_info_key_header_file),
    "interface/macro": ("macro", _augment_interface, _ref_function,
                        _ref_disabled_function, _link_function,
                        _name_info_key_define),
    "interface/register-block": ("register block", _augment_interface,
                                 _ref_name, _ref_disabled_name, _link_name,
                                 _name_info_key_type),
    "interface/struct": ("structure", _augment_interface, _ref_name,
                         _ref_disabled_name, _link_name, _name_info_key_type),
    "interface/typedef": ("type definition", _augment_interface, _ref_name,
                          _ref_disabled_name, _link_name, _name_info_key_type),
    "interface/union": ("union", _augment_interface, _ref_name,
                        _ref_disabled_name, _link_name, _name_info_key_type),
    "interface/unspecified-define": ("define", _augment_interface, _ref_name,
                                     _ref_disabled_name, _link_name,
                                     _name_info_key_define),
    "interface/unspecified-enum":
    ("enumeration", _augment_interface, _ref_unspecified_type,
     _ref_disabled_name, _link_unspecified_type, _name_info_key_enum),
    "interface/unspecified-enumerator":
    ("enumerator", _augment_interface, _ref_name, _ref_disabled_name,
     _link_name, _name_info_key_enumerator),
    "interface/unspecified-function":
    ("directive", _augment_interface, _ref_function, _ref_disabled_function,
     _link_function, _name_info_key_function),
    "interface/unspecified-group": ("group", _augment_interface, _ref_name,
                                    _ref_disabled_name, _link_name,
                                    _name_info_key_group),
    "interface/unspecified-header-file":
    ("header file", _augment_interface, _ref_header_file,
     _ref_disabled_header_file, _link_header_file, _name_info_key_header_file),
    "interface/unspecified-macro": ("macro", _augment_interface, _ref_function,
                                    _ref_disabled_function, _link_function,
                                    _name_info_key_define),
    "interface/unspecified-object": ("object", _augment_interface, _ref_name,
                                     _ref_disabled_name, _link_name,
                                     _name_info_key_object),
    "interface/unspecified-struct":
    ("type", _augment_interface, _ref_unspecified_type, _ref_disabled_name,
     _link_unspecified_type, _name_info_key_type),
    "interface/unspecified-typedef": ("type definition", _augment_interface,
                                      _ref_name, _ref_disabled_name,
                                      _link_name, _name_info_key_type),
    "interface/unspecified-union": ("type", _augment_interface,
                                    _ref_unspecified_type, _ref_disabled_name,
                                    _link_unspecified_type,
                                    _name_info_key_type),
    "interface/variable": ("object", _augment_interface, _ref_name,
                           _ref_disabled_name, _link_name,
                           _name_info_key_object),
    "memory-benchmark": ("memory benchmark", _augment_memory_benchmark,
                         _ref_spec, _ref_disabled_spec, _link_default,
                         _name_info_key_default),
    "requirement/functional/action": ("action requirement",
                                      _augment_requirement_self_test,
                                      _ref_spec, _ref_disabled_spec,
                                      _link_default, _name_info_key_default),
    "requirement/functional/capability": ("capability requirement",
                                          _augment_requirement, _ref_spec,
                                          _ref_disabled_spec, _link_default,
                                          _name_info_key_default),
    "requirement/functional/interface-define-not-defined":
    ("interface define requirement", _augment_requirement, _ref_spec,
     _ref_disabled_spec, _link_default, _name_info_key_default),
    "requirement/functional/fatal-error": ("fatal error", _augment_requirement,
                                           _ref_spec, _ref_disabled_spec,
                                           _link_default,
                                           _name_info_key_default),
    "requirement/functional/function": ("function requirement",
                                        _augment_requirement, _ref_spec,
                                        _ref_disabled_spec, _link_default,
                                        _name_info_key_default),
    "requirement/non-functional/design": ("design requirement",
                                          _augment_requirement, _ref_spec,
                                          _ref_disabled_spec, _link_default,
                                          _name_info_key_default),
    "requirement/non-functional/design-group":
    ("design group", _augment_design_group, _ref_spec, _ref_disabled_spec,
     _link_design_group, _name_info_key_group),
    "requirement/non-functional/design-target":
    ("design target", _augment_requirement, _ref_spec, _ref_disabled_spec,
     _link_default, _name_info_key_default),
    "requirement/non-functional/interface": ("interface requirement",
                                             _augment_requirement, _ref_spec,
                                             _ref_disabled_spec, _link_default,
                                             _name_info_key_default),
    "requirement/non-functional/interface-requirement":
    ("interface requirement", _augment_interface_requirement, _ref_spec,
     _ref_disabled_spec, _link_default, _name_info_key_default),
    "requirement/non-functional/performance":
    ("performance requirement", _augment_requirement, _ref_spec,
     _ref_disabled_spec, _link_default, _name_info_key_default),
    "requirement/non-functional/performance-runtime":
    ("runtime performance requirement", _augment_requirement_self_test,
     _ref_spec, _ref_disabled_spec, _link_default, _name_info_key_default),
    "requirement/non-functional/performance-runtime-environment":
    ("runtime performance measurement environment", _augment_requirement,
     _ref_spec, _ref_disabled_spec, _link_default, _name_info_key_default),
    "requirement/non-functional/quality": ("quality requirement",
                                           _augment_requirement, _ref_spec,
                                           _ref_disabled_spec, _link_default,
                                           _name_info_key_default),
    "runtime-measurement-test": ("runtime measurement test",
                                 _augment_test_case, _ref_spec,
                                 _ref_disabled_spec, _link_default,
                                 _name_info_key_default),
    "spec": ("specification specification", _augment_default, _ref_spec,
             _ref_disabled_spec, _link_default, _name_info_key_default),
    "test-case": ("test case", _augment_test_case, _ref_spec,
                  _ref_disabled_spec, _link_default, _name_info_key_default),
    "test-suite": ("test suite", _augment_test_suite, _ref_spec,
                   _ref_disabled_spec, _link_default, _name_info_key_default),
    "validation/by-analysis": ("validation by analysis",
                               _augment_other_validation, _ref_spec,
                               _ref_disabled_spec, _link_default,
                               _name_info_key_default),
    "validation/by-inspection": ("validation by inspection",
                                 _augment_other_validation, _ref_spec,
                                 _ref_disabled_spec, _link_default,
                                 _name_info_key_default),
    "validation/by-review-of-design": ("validation by review of design",
                                       _augment_other_validation, _ref_spec,
                                       _ref_disabled_spec, _link_default,
                                       _name_info_key_default),
}


def get_kind(item: Item) -> str:
    """ Returns the item kind. """
    return _ITEM_SPECIFICS[item.type][0]


def _get_group_items(name_info: Dict[str, Any], spec: RTEMSItemCache,
                     info: Dict[str, Any]) -> List[Item]:
    items = []
    try:
        for group in info["groups"]:
            name = f"group/{group}"
            try:
                items.append(spec.name_to_item[name])
            except KeyError:
                items.extend(_get_group_items(name_info, spec,
                                              name_info[name]))
    except KeyError:
        doxygen = info["doxygen"]
        if doxygen.startswith("@file"):
            spec.item.view.setdefault("files-without-group",
                                      set()).add(doxygen[6:])
        if doxygen.startswith("@addtogroup"):
            spec.item.view.setdefault("groups-without-items",
                                      set()).add(doxygen[11:])
    return items


def _get_associated_items(name_info: Dict[str, Any], spec: RTEMSItemCache,
                          name: str, info: Dict[str, Any]) -> List[Item]:
    item = spec.name_to_item.get(name, None)
    if item is not None:
        return [item]
    groups = info.get("groups", None)
    if groups is not None:
        return _get_group_items(name_info, spec, info)
    if name.startswith("file/"):
        file_info = info
    else:
        try:
            file_info = name_info[f"file/{info['file']}"]
        except KeyError as err:
            raise ValueError(
                f"'{name}' is not associated with a file: {info}") from err
    return _get_group_items(name_info, spec, file_info)


def _augment_name_info_default(name_info: Dict[str, Any], spec: RTEMSItemCache,
                               name: str, info: Dict[str, Any]) -> None:
    for variant in itertools.chain([info], info.get("variants", [])):
        items = _get_associated_items(name_info, spec, name, variant)
        variant["items"] = set(items)


def _augment_name_info_group(name_info: Dict[str, Any], spec: RTEMSItemCache,
                             name: str, info: Dict[str, Any]) -> None:
    try:
        items = [spec.name_to_item[name]]
    except KeyError:
        items = _get_group_items(name_info, spec, info)
    info["items"] = set(items)


def _augment_name_info_nop(_name_info: Dict[str, Any], _spec: RTEMSItemCache,
                           _name: str, _info: Dict[str, Any]) -> None:
    pass


_AUGMENT_NAME_INFO = {
    "dir": _augment_name_info_nop,
    "file-key": _augment_name_info_nop,
    "group": _augment_name_info_group,
    "groups": _augment_name_info_nop
}


def _check_mapping_completeness(spec: RTEMSItemCache) -> None:
    without = spec.item.view.get("files-without-group")
    if without:
        raise ValueError(f"files without group: {' '.join(sorted(without))}")
    without = spec.item.view.get("groups-without-items")
    if without:
        raise ValueError(f"groups without items: {' '.join(sorted(without))}")


def _augment_name_info(name_info: Dict[str, Any],
                       spec: RTEMSItemCache) -> None:
    # Associate files and groups with groups
    for group in name_info["groups/all"]:
        info = name_info[f"group/{group}"]
        files: List[str] = []
        for key in sorted(info["file-keys"]):
            file_key = name_info[f"file-key/{key}"]
            files.append(file_key)
            file_info = name_info[f"file/{file_key}"]
            file_info.setdefault("groups", []).append(group)
        info["files"] = files
        for key in sorted(info["subgroups"]):
            group_info = name_info[f"group/{key}"]
            group_info.setdefault("groups", []).append(group)

    for name, info in name_info.items():
        if isinstance(info, dict) and "instances" in info:
            instances = info["instances"]
            if name.startswith("define/CONFIGURE"):
                for anchorfile, instance in instances.items():
                    if next(iter(
                            instance["groups"])).startswith("RTEMSApplConfig"):
                        first = anchorfile
                        break
                else:
                    first = sorted(instances.keys())[0]
            else:
                first = sorted(instances.keys())[0]
            info.update(instances[first])
            del instances[first]
            info["variants"] = list(instances.values())
            del info["instances"]
    for name, info in name_info.items():
        _AUGMENT_NAME_INFO.get(name[0:name.find("/")],
                               _augment_name_info_default)(name_info, spec,
                                                           name, info)
    _check_mapping_completeness(spec)


class LinkHub(BuildItem):
    """ Provides links from items to documents. """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        self.name_info = {}  # type: Dict[str, Any]
        for link in self.item.links_to_parents("document"):
            target = self.director[link.item.uid]
            assert isinstance(target, DirectoryState)
            name = f"dir/{link['name']}"
            self.name_info[name] = os.path.normpath(
                os.path.join(target.directory, link["directory"]))
        try:
            tagfile = self.input("tagfile")
        except KeyError as err:
            if "RTEMS_QUAL" in self.enabled_set:
                raise ValueError(
                    f"{self.uid}: there is no associated tagfile") from err
        else:
            assert isinstance(tagfile, DirectoryState)
            _gather_name_info(self.name_info, tagfile.file)
            spec = self.input("spec")
            assert isinstance(spec, RTEMSItemCache)
            _augment_name_info(self.name_info, spec)
            for item_2 in spec.item_cache.values():
                if not item_2.enabled:
                    continue
                _ITEM_SPECIFICS.get(item_2.type,
                                    _ITEM_DEFAULT)[1](item_2, self.name_info)

    def run(self) -> None:
        self.description.add("""Provide documentation links for specification
items.""")

    def get_sdd_link(self, kind: str, name: str) -> str:
        """ Return the SDD link for the name by kind. """
        elem_info = self.name_info[f"{kind}/{name}"]
        return f"`{elem_info['name']} <{elem_info['link']}>`__"

    def get_file_sdd_link(self, path: str) -> str:
        """ Return the SDD link for the file path. """
        return self.get_sdd_link("file", path)

    def get_function_sdd_link(self, name: str) -> str:
        """ Return the SDD link for the function with the name. """
        return self.get_sdd_link("function", name)

    def get_any_sdd_link(self, name: str) -> str:
        """ Return an SDD link for the name. """
        for kind in _ELEMENT_KINDS:
            elem_info = self.name_info.get(f"{kind}/{name}")
            if elem_info is not None:
                break
        if elem_info is None:
            raise ValueError("there is no SDD element with this name: {name}")
        return f"`{elem_info['name']} <{elem_info['link']}>`__"


class SpecMapper(BuildItemMapper):
    """ Item mapper for specifications. """

    def __init__(self, whoami: str, build_item: BuildItem, item: Item):
        super().__init__(item)
        self._build_item = build_item
        self._whoami = whoami
        self.add_get_value("interface/appl-config-option/feature-enable:/name",
                           self._get_value_link)
        self.add_get_value("interface/appl-config-option/feature:/name",
                           self._get_value_link)
        self.add_get_value("interface/appl-config-option/initializer:/name",
                           self._get_value_link)
        self.add_get_value("interface/appl-config-option/integer:/name",
                           self._get_value_link)
        self.add_get_value("interface/define:/name", self._get_value_link)
        self.add_get_value("interface/enum:/name", self._get_value_link)
        self.add_get_value("interface/enumerator:/name", self._get_value_link)
        self.add_get_value("interface/function:/name", self._get_value_link)
        self.add_get_value("interface/group:/name", self._get_value_link)
        self.add_get_value("interface/header-file:/path", self._get_value_link)
        self.add_get_value("interface/macro:/name", self._get_value_link)
        self.add_get_value("interface/struct:/name", self._get_value_link)
        self.add_get_value("interface/typedef:/name", self._get_value_link)
        self.add_get_value("interface/union:/name", self._get_value_link)
        self.add_get_value("interface/unspecified-define:/name",
                           self._get_value_link)
        self.add_get_value("interface/unspecified-enumerator:/name",
                           self._get_value_link)
        self.add_get_value("interface/unspecified-function:/name",
                           self._get_value_link)
        self.add_get_value("interface/unspecified-group:/name",
                           self._get_value_link)
        self.add_get_value("interface/unspecified-enum:/name",
                           self._get_value_link)
        self.add_get_value("interface/unspecified-struct:/name",
                           self._get_value_link)
        self.add_get_value("interface/unspecified-typedef:/name",
                           self._get_value_link)
        self.add_get_value("interface/unspecified-union:/name",
                           self._get_value_link)
        self.add_get_value("interface/variable:/name", self._get_value_link)
        self.add_get_value("requirement/non-functional/design-group:/name",
                           self._get_value_link)
        for type_name in self.item.cache.items_by_type.keys():
            self.add_get_value(f"{type_name}:/spec", self._get_value_link)
        for kind in ("define", "enum", "enumerator", "file", "function",
                     "group", "object", "type"):
            self.add_default_get_value(
                f"sdd-{kind}", functools.partial(self._get_value_sdd_link,
                                                 kind))
        self.add_default_get_value("sdd", self._get_value_sdd_link_unique)

    def make_reference(self, item: Item, name: Optional[str] = None) -> str:
        """
        Make a document internal reference for the item with the optional
        name.

        When no name is given, then the link name is item.spec_2.
        """
        assert item.enabled
        if name is None:
            name = item.spec_2
        return f":ref:`{name} <{spec_label(item)}>`"

    def get_link(self, item: Item, document_key: None | str = None) -> str:
        if not item.enabled:
            return _ITEM_SPECIFICS.get(item.type, _ITEM_DEFAULT)[3](item)
        try:
            if document_key not in item.view["document-links"]:
                document_key = item.view["default-document"]
            if document_key == self._whoami:
                document_key = "self"
            return item.view["document-links"][document_key]
        except KeyError:
            assert "RTEMS_QUAL" not in item.cache.enabled_set
            return _spec_link(item, _EB_PKG_SUPPORT)

    def _get_value_link(self, ctx: ItemGetValueContext) -> str:
        return self.get_link(ctx.item)

    def _get_value_sdd_link(self, kind: str, ctx: ItemGetValueContext) -> str:
        link_hub_uid = self._build_item.component.item.child("link-hub").uid
        link_hub = self._build_item.director[link_hub_uid]
        assert isinstance(link_hub, LinkHub)
        assert ctx.args
        return link_hub.get_sdd_link(kind, ctx.args)

    def _get_value_sdd_link_unique(self, ctx: ItemGetValueContext) -> str:
        link_hub_uid = self._build_item.component.item.child("link-hub").uid
        link_hub = self._build_item.director[link_hub_uid]
        assert isinstance(link_hub, LinkHub)
        assert ctx.args
        return link_hub.get_any_sdd_link(ctx.args)



# SPDX-License-Identifier: BSD-2-Clause
"""
Provides a command line interface to update the test timeouts using the
specified test reports.
"""

# Copyright (C) 2024, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import argparse
import dataclasses
import datetime
import json
import logging
from pathlib import Path
import sys

from specitems import (Item, ItemCache, ItemCacheConfig,
                       get_item_cache_arguments)


@dataclasses.dataclass
class _LastUpdate:
    old: datetime.datetime
    new: datetime.datetime


def _update_timeouts(args: argparse.Namespace, report_path: str,
                     last_update: _LastUpdate, timeouts: dict[str, list[int]],
                     reports: list) -> None:
    error_factor = args.error_factor
    warning_factor = args.warning_factor
    minimum_timeout = args.minimum_timeout
    for report in reports:
        name = Path(report["executable"]).name
        new_duration = report.get("execution-duration-in-seconds",
                                  report.get("duration"))
        if not new_duration:
            logging.debug("%s: %s: has no execution duration", report_path,
                          name)
            continue
        update_time: str | None = report.get("start-time")
        if update_time is not None:
            update_datetime = datetime.datetime.fromisoformat(update_time)
            if update_datetime <= last_update.old:
                logging.debug("%s: %s: skip out of date result", report_path,
                              name)
                continue
            last_update.new = max(last_update.new, update_datetime)
        durations = timeouts.setdefault(name, [])
        try:
            maximum = max(durations)
        except ValueError:
            pass
        else:
            if new_duration > error_factor * maximum + minimum_timeout:
                logging.error(
                    "%s: %s: duration %s is greater than %s * %s + %s",
                    report_path, name, new_duration, error_factor, maximum,
                    minimum_timeout)
                continue
            if new_duration > warning_factor * maximum + minimum_timeout:
                logging.warning(
                    "%s: %s: duration %s is greater than %s * %s + %s",
                    report_path, name, new_duration, warning_factor, maximum,
                    minimum_timeout)
        logging.debug("%s: %s: add duration: %s", report_path, name,
                      new_duration)
        durations.append(new_duration)


def _prepare_timeouts(item: Item, report_path: str, data: dict, reset: bool,
                      reset_done: set[str]) -> dict[str, list[int]]:
    timeout_key = data["timeout-key"]
    logging.info("%s: timeout key: %s", report_path, timeout_key)
    reset_key = f"{item.uid} {timeout_key}"
    if reset and reset_key not in reset_done:
        reset_done.add(reset_key)
        timeouts = {}
    else:
        timeouts = item["timeouts"].get(timeout_key, {})
    item["timeouts"][timeout_key] = timeouts
    return timeouts


def _get_arguments(argv: list[str]) -> argparse.Namespace:

    def _add_arguments(parser):
        parser.add_argument('--dry-run',
                            action="store_true",
                            default=False,
                            help="do not save timeout items")
        parser.add_argument('--reset',
                            action="store_true",
                            help="reset the timeouts")
        parser.add_argument(
            '--test-timeouts',
            type=str,
            default="test-timeouts",
            help="the test timeouts item UID basename (default: test-timeouts)"
        )
        parser.add_argument(
            '--minimum-timeout',
            type=float,
            default=10.0,
            help="the minimum timeout used by the test runner (default: 10.0)")
        parser.add_argument(
            '--warning-factor',
            type=float,
            default=1.2,
            help="a new duration greater than the factor times the current "
            "maximum duration plus the minimum timeout "
            "is a warning (default: 1.2)")
        parser.add_argument(
            '--error-factor',
            type=float,
            default=1.9,
            help="a new duration greater than the factor times the current "
            "maximum duration plus the minimum timeout "
            "is an error (default: 1.9)")
        parser.add_argument("reports",
                            metavar="REPORT",
                            nargs="+",
                            help="a test report JSON file")

    return get_item_cache_arguments(argv,
                                    description=cliupdatetimeouts.__doc__,
                                    add_arguments=(_add_arguments, ))


def cliupdatetimeouts(argv: list[str] = sys.argv) -> None:
    """
    Update the test timeouts using the specified test reports.

    The target UID and the timeout key are obtained from the test report.
    """
    args = _get_arguments(argv[1:])
    cache_config = ItemCacheConfig(paths=args.spec_directories,
                                   cache_directory=args.cache_directory,
                                   initialize_links=False,
                                   resolve_proxies=False)
    item_cache = ItemCache(cache_config)
    reset_done: set[str] = set()
    last_updates: dict[str, _LastUpdate] = {}
    for report_path in args.reports:
        logging.info("%s: evaluate reports", report_path)
        with open(report_path, "r", encoding="utf-8") as src:
            data = json.load(src)
        try:
            target = data["target"]
        except KeyError:
            logging.warning("%s: report has no target attribute", report_path)
            continue
        uid = f"{target}/{args.test_timeouts}"
        logging.info("%s: test timeouts item UID: %s", report_path, uid)
        item = item_cache[uid]
        time_of_last_update = datetime.datetime.fromisoformat(
            item.get("time-of-last-update", "1970-01-01"))
        last_update = last_updates.setdefault(
            uid, _LastUpdate(time_of_last_update, time_of_last_update))
        timeouts = _prepare_timeouts(item, report_path, data, args.reset,
                                     reset_done)
        _update_timeouts(args, report_path, last_update, timeouts,
                         data["reports"])
        if not args.dry_run:
            item["time-of-last-update"] = last_updates[uid].new.isoformat()
            item.save()



# SPDX-License-Identifier: BSD-2-Clause
""" Builds test plans documents. """

# Copyright (C) 2022, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from specitems import (Item, ItemGetValueContext, SphinxContent, to_camel_case)
from specware import gather_benchmarks_and_test_suites, gather_test_cases

from .pkgitems import PackageBuildDirector
from .linkhub import LinkHub, spec_label
from .specdocbuilder import SpecDocumentBuilder
from .testrunner import TestRunner


def _get_test_case_function(item: Item, ident: str) -> str:
    if item["test-header"] is None:
        return f"T_case_body_{ident}"
    return f"{ident}_Run"


def _add_test_results(content: SphinxContent, kind: str, item: Item) -> None:
    content.add_rubric("TEST RESULTS:")
    lines: list[str] = []
    for test_results in item.view.get("test-results", {}).values():
        for data in test_results:
            config_data = data["config"]
            target_data = config_data["target"]
            name = f"Target - {target_data['name']}"
            name = f"{name} / Configuration - {config_data['name']}"
            test_suite = data.get("test-suite", None)
            if test_suite is not None:
                spec = item.cache[test_suite['uid']].spec_2
                name = f"{name} / Test suite - {spec}"
            lines.append(f"`{name} <{data['link']}>`__")
    content.add_list(lines,
                     f"""For this {kind}, the following test results are
available:""",
                     empty="There are no test results available.")


class TestPlanBuilder(SpecDocumentBuilder):
    """ Builds a test plan document such as SVS or SUITP. """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        my_type = self.item.type
        self.mapper.add_get_value(f"{my_type}:/test-suites",
                                  self._get_test_suites)
        self.mapper.add_get_value(f"{my_type}:/test-cases",
                                  self._get_test_cases)
        self.mapper.add_get_value(f"{my_type}:/test-procedures",
                                  self._get_test_procedures)
        self.mapper.add_get_value(f"{my_type}:/other-validations",
                                  self._get_other_validations)
        self.mapper.add_get_value(f"{my_type}:/not-validated-by-test",
                                  self._get_not_validated_by_test)
        self._test_suites: list[Item] = []
        self._documenter = {
            "requirement/functional/action": self._document_action,
            "requirement/non-functional/performance-runtime":
            self._document_performance_runtime,
            "runtime-measurement-test":
            self._document_runtime_measurement_request,
            "test-case": self._document_test_case
        }

    def _get_link_hub(self) -> LinkHub:
        link_hub = self.input("link-hub")
        assert isinstance(link_hub, LinkHub)
        return link_hub

    def _gather_test_suites(self) -> list[Item]:
        test_suites = self._test_suites
        if not test_suites:
            for parent in self.inputs("test-suites"):
                gather_benchmarks_and_test_suites(parent.item, test_suites)
            test_suites.sort()
        return test_suites

    def _add_test_suite(self, content: SphinxContent, item: Item) -> None:
        with content.directive("raw", "latex"):
            content.add("\\clearpage")
        with content.section(item.spec, label=spec_label(item)):
            with content.section("General"):
                content.add_rubric("DESCRIPTION:")
                self.wrap(content, item, item["test-brief"])
                self.wrap(content, item, item["test-description"])
                if item.type != "memory-benchmark":
                    _add_test_results(content, "test suite", item)
                self.add_item_changes(content, item)
            with content.section("Features to be tested"):
                gathered_test_cases: list[Item] = []
                gather_test_cases(item, gathered_test_cases)
                test_cases = sorted(set(gathered_test_cases))
                count = len(test_cases)
                if item.type == "memory-benchmark":
                    assert count == 0
                    content.add(
                        self.mapper.substitute("""Memory usage
benchmarks contain no test cases.  They are not executed on the
${/glossary/target:/term}.  Instead the generated ${/glossary/elf:/term} file
is analysed to gather the memory usage of sections and data structures.  The
analysis results are presented in the
${/pkg/deployment/doc-package-manual:/cite-long}."""))
                elif count == 1:
                    link = self.mapper.make_reference(test_cases[0])
                    content.add("The features to be tested are defined by "
                                f"the test case {link}.")
                elif count > 1:
                    prologue = ("The features to be tested are defined by "
                                "the following test cases:")
                    content.add_list([
                        self.mapper.make_reference(item_2)
                        for item_2 in test_cases
                    ], prologue)
                else:
                    content.add(
                        "The test suite contains no specified test cases.")
            with content.section("Approach refinements"):
                content.add(f"""There are no approach refinements
necessary.  The test suite is implemented in the file
{self._get_link_hub().get_file_sdd_link(item['test-target'])}.""")

    def _get_test_suites(self, _ctx: ItemGetValueContext) -> str:
        content = SphinxContent(section_level=2)
        for item in self._gather_test_suites():
            with self.mapper.scope(item):
                self._add_test_suite(content, item)
        return content.join()

    def _add_test_case_validations(self, content: SphinxContent, item: Item,
                                   links: list[dict[str,
                                                    str]], what: str) -> bool:
        specs: list[str] = []
        for link in links:
            if link["role"] != "validation":
                continue
            item_2 = item.map(link["uid"])
            if not item_2.enabled:
                continue
            specs.append(self.mapper.get_link(item_2))
        count = len(specs)
        if count == 0:
            return True
        if count == 1:
            validations = specs[0]
        elif count == 2:
            validations = f"{specs[0]} and {specs[1]}"
        else:
            specs[-1] = f"and {specs[-1]}"
            validations = ", ".join(specs)
        content.add(f"{what} {validations}.")
        return False

    def _document_action(self, content: SphinxContent, link_hub: LinkHub,
                         item: Item, ident: str) -> None:
        test_case_function = _get_test_case_function(item, ident)
        content.add(f"""This test case validates all state transitions
specified by the action requirement {self.mapper.get_link(item)}.  The
transition map is validated by the function
{link_hub.get_function_sdd_link(test_case_function)} contained in the
file {link_hub.get_file_sdd_link(item['test-target'])}.""")

    def _document_performance_runtime(self, content: SphinxContent,
                                      link_hub: LinkHub, item: Item,
                                      ident: str) -> None:
        context = item.parent("runtime-measurement-request")
        test_case_function = f"{ident}_Body"
        content.add(f"""This test case performs a performance runtime
measurement request which is carried out by
{self.mapper.make_reference(context)}.  It produces the runtime measurements
required by the runtime performance requirement {self.mapper.get_link(item)}.
It is implemented by the function
{link_hub.get_function_sdd_link(test_case_function)} contained in the
file {link_hub.get_file_sdd_link(context['test-target'])}.""")

    def _document_runtime_measurement_request(self, content: SphinxContent,
                                              link_hub: LinkHub, item: Item,
                                              ident: str) -> None:
        self.wrap(content, item, item["test-brief"])
        content.gap = False
        content.add_list(
            sorted([
                self.mapper.make_reference(item_2)
                for item_2 in item.children("runtime-measurement-request")
            ]), "The following runtime measurement requests are carried out:")
        content.gap = self._add_test_case_validations(
            content, item, item["links"], "This test case validates")
        test_case_function = f"T_case_body_{ident}"
        content.add(f"""This test case is implemented by the
function {link_hub.get_function_sdd_link(test_case_function)} contained
in the file {link_hub.get_file_sdd_link(item['test-target'])}.""")

    def _document_test_case(self, content: SphinxContent, link_hub: LinkHub,
                            item: Item, ident: str) -> None:
        self.wrap(content, item, item["test-brief"])
        self.wrap(content, item, item["test-description"])
        content.append("The following test case actions are carried out:")
        for index, action in enumerate(item["test-actions"]):
            with content.list_item(
                    self.mapper.substitute(action["action-brief"])):
                for check in action["checks"]:
                    with content.list_item(
                            self.mapper.substitute(check["brief"])):
                        content.gap = False
                        self._add_test_case_validations(
                            content, item, check["links"], "This validates")
                content.gap = bool(action["checks"])
                content.gap = self._add_test_case_validations(
                    content, item, action["links"], "This action validates")
                action_function = f"{ident}_Action_{index}"
                content.add(f"""This action is implemented by
the function {link_hub.get_function_sdd_link(action_function)}.""")
        test_case_function = _get_test_case_function(item, ident)
        content.add(f"""This test case is implemented by the
function {link_hub.get_function_sdd_link(test_case_function)} contained
in the file {link_hub.get_file_sdd_link(item['test-target'])}.""")

    def _add_test_suite_memberships(self, content: SphinxContent,
                                    item: Item) -> None:
        test_suites: list[str] = [
            self.mapper.make_reference(test_suite)
            for test_suite in item.parents("test-case")
        ]
        if not test_suites:
            return
        content.add_rubric("TEST SUITES:")
        if len(test_suites) == 1:
            content.add("This test case is contained in the "
                        f"{test_suites[0]} test suite.")
        else:
            content.add_list(
                test_suites,
                "This test case is contained in the following test suites:")

    def _add_test_case(self, content: SphinxContent, item: Item) -> None:
        link_hub = self._get_link_hub()
        with content.directive("raw", "latex"):
            content.add("\\clearpage")
        with content.section(item.spec, label=spec_label(item)):
            with content.section("General"):
                content.add_rubric("DESCRIPTION:")
                ident = to_camel_case(item.uid[1:])
                self._documenter[item.type](content, link_hub, item, ident)
                self._add_test_suite_memberships(content, item)
                _add_test_results(content, "test case", item)
                self.add_item_changes(content, item)
            with content.section("Input specifications"):
                content.add("""The test case starts execution in the
system state defined by the test suite runner and the test suite configuration.
All other inputs required by the test case are produced by the test case code
itself.""")
            with content.section("Output specifications"):
                content.add("For the output specifications see section "
                            ":ref:`TestPassFailCriteria`.")
            with content.section("Test pass - fail criteria"):
                content.add("For the test pass - fail criteria see section "
                            ":ref:`TestPassFailCriteria`.")
            with content.section("Environmental needs"):
                content.add("There are no specific environmental needs.")
            with content.section("Special procedure constraints"):
                content.add(
                    "There are no special procedure constraints applicable.")
            with content.section("Interface dependencies"):
                content.add(
                    "There are no specific interface dependencies present.")

    def _get_test_cases(self, _ctx: ItemGetValueContext) -> str:
        content = SphinxContent(section_level=2)
        test_cases: list[Item] = []
        for item in self._gather_test_suites():
            gather_test_cases(item, test_cases)
        for item in sorted(set(test_cases)):
            with self.mapper.scope(item):
                self._add_test_case(content, item)
        return content.join()

    def _get_test_procedures(self, _ctx: ItemGetValueContext) -> str:
        test_runners: list[TestRunner] = []
        for item in self.component.item.parents("test-runner"):
            test_runner = self.director[item.uid]
            test_runners.append(test_runner)
        content = SphinxContent(section_level=2)
        for test_runner in sorted(test_runners):
            label = spec_label(test_runner.item)
            with test_runner.component_scope(self.component):
                with content.section(test_runner["name"], label=label):
                    content.add(test_runner.describe())
        return content.join()

    def _add_other_validation(self, content: SphinxContent,
                              method: str) -> None:
        items = self.spec.get_related_items_by_type(
            f"validation/by-{method.replace(' ', '-')}")
        if items:
            with content.section(f"Validation by {method}"):
                content.add(
                    "This section lists validation evidence obtained by "
                    f"using the {method} validation method.")
                for item in items:
                    self.add_item(content, item)

    def _get_other_validations(self, _ctx: ItemGetValueContext) -> str:
        content = SphinxContent(section_level=2)
        for method in ("analysis", "inspection", "review of design"):
            self._add_other_validation(content, method)
        return content.join()

    def _get_not_validated_by_test(self, _ctx: ItemGetValueContext) -> str:
        content = SphinxContent(section_level=2)
        types = [
            "validation/by-analysis", "validation/by-inspection",
            "validation/by-review-of-design"
        ]
        items: list[Item] = []
        for item in self.spec.get_related_items_by_type(types):
            items.extend(parent for parent in item.parents("validation"))
        content.add_list([
            self.mapper.get_link(item, document_key="default")
            for item in sorted(set(items))
        ], "The following items are not specifically validated by a test:")
        return content.join()
